/********************************************************************
  APG - an ABNF Parser Generator
  Copyright  (C) 2007 Coast to Coast Research, Inc.
 
  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 2 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program.  If not, see
  <http://www.gnu.org/licenses/old-licenses/gpl-2.0.html>
  or write to the Free Software Foundation, Inc.,
  51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

    author: Lowell Thomas
            lowell@coasttocoastresearch.com
            http://www.coasttocoastresearch.com

   created: 8/25/2007 10:11
  filename: D:\Projects\Apgv5_0\ApgLib\Memory.c
   purpose: Source code file for the memory component functions.
  
*********************************************************************/
#include "Apg.h"
#include "Private.h"

#define MEM_ID_ALLOC    0
#define MEM_ID_REALLOC  1
#define MEM_ID_FREE     2
static void vMemStats(APG_MEM_STATS* spStats, ulong ulID, APG_MEM_CELL* spIn, APG_MEM_CELL* spOut);
static void vActivePush(APG_MEM_CTX* spCtx, APG_MEM_CELL* spCellIn);
static void vActivePop(APG_MEM_CTX* spCtx, APG_MEM_CELL* spCellIn);

//////////////////////////////////////////////////////////////////////////
// MEMORY MANAGEMENT COMPONENT FUNCTIONS
//   - Lowest-level APG component
//   - Does NO error reporting other than returning a NULL pointer
//     as an allocation failure indicator.
//   - keeps a linked list of memory allocations and does garbage
//     collection on destruction
//   - Does debug-only reporting of invalid input argument errors.
//////////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////////
// DESCRIPTION: memory component constructor
//
// FUNCTION: void* vpMemCtorA
//
//      ARG: 1) PFN_ALLOCATOR pfnAllocator
//              : pointer to the memory allocator
//
//      ARG: 2) PFN_DEALLOCATOR pfnDeAllocator
//              : pointer to the memory deallocator
//
//   RETURN: returns pointer to a context
//
//////////////////////////////////////////////////////////////////////////
void*   vpMemCtorA(PFN_ALLOCATOR pfnAllocator, PFN_DEALLOCATOR pfnDeAllocator)
{
  void*           vpReturn = NULL;
  ulong           ulSize;
  APG_MEM_CTX*    spCtx;
  APG_ERR_CTX*    spErrCtx;
  DASSERT(pfnAllocator && pfnDeAllocator);

  // get a context using custom allocator
  ulSize = sizeof(APG_MEM_CTX) + sizeof(APG_ERR_CTX) + (APG_ERR_DEFAULT * sizeof(APG_ERR));
  spCtx = (APG_MEM_CTX*)pfnAllocator(ulSize);
  if(spCtx)
  {
    // initialize the memory context
    memset((void*)spCtx, 0, ulSize);
    spCtx->pfnAllocator = pfnAllocator;
    spCtx->pfnDeAllocator = pfnDeAllocator;

    // initialize the err context
    spErrCtx = (APG_ERR_CTX*)(spCtx + 1);
    spErrCtx->spErrors = (APG_ERR*)(spErrCtx + 1);
    spErrCtx->ulAvailable = ulSize;
    spErrCtx->ulUsed = 0;
    spErrCtx->ulTotal = 0;

    // set valid
    spCtx->spErrCtx = spErrCtx;
    spCtx->vpValidate = (void*)spCtx;
    vpReturn = (void*)spCtx;
  }

  return vpReturn;
}

// memory component constructor using malloc() and free() allocation
//////////////////////////////////////////////////////////////////////////
// DESCRIPTION: memory component constructor
// uses malloc() and free() memory (de-)allocation
//
// FUNCTION: void* vpMemCtor
//
//   RETURN: void
//
//////////////////////////////////////////////////////////////////////////
void* vpMemCtor()
{
  return vpMemCtorA(malloc, free);
}

//////////////////////////////////////////////////////////////////////////
// DESCRIPTION: memory component destructor
//
// FUNCTION: void vMemDtor
//
//      ARG: 1) void* vpCtx
//              : pointer to a memory context previously returned
//                from vpMemCtor() or vpMemCtorA()
//                NULL pointers are ignored, no action and no error
//
//   RETURN: void
//
//////////////////////////////////////////////////////////////////////////
void   vMemDtor(void* vpCtx)
{
  APG_MEM_CTX* spCtx = (APG_MEM_CTX*)vpCtx;
  if(vpCtx)
  {
    DASSERT(spCtx->vpValidate == (void*)spCtx);
    
    // release all active cells
    vMemFreeToCheckPoint(vpCtx, 0);
    
    // release this context
    spCtx->pfnDeAllocator(vpCtx);
  }
}

//////////////////////////////////////////////////////////////////////////
// DESCRIPTION: allocates memory
//
// FUNCTION: void* vpMemAlloc
//
//      ARG: 1) void* vpCtx
//              : pointer to a memory context previously returned
//                from vpMemCtor() or vpMemCtorA()
//
//      ARG: 2) ulong ulBytes
//              : number of bytes of memory to allocate
//
//   RETURN: returns a pointer to the allocated memory
//            returns NULL if memory allocation fails
//            and records the error with the error component
//
//////////////////////////////////////////////////////////////////////////
void* vpMemAlloc(void* vpCtx, ulong ulBytes)
{
  void*         vpReturn = NULL;
  APG_MEM_CTX*  spCtx = (APG_MEM_CTX*)vpCtx;
  APG_MEM_CELL* spCell;
  DASSERT(spCtx && spCtx->vpValidate == (void*)spCtx);

  // allocate the memory
  spCell = (APG_MEM_CELL*)spCtx->pfnAllocator(ulBytes + sizeof(APG_MEM_CELL));
  if(spCell)
  {
    spCell->ulSize = ulBytes;
    
    // push it on the active list
    vActivePush(spCtx, spCell);
    APG_MEMSTATS(&spCtx->sStats, MEM_ID_ALLOC, spCell, 0);

    // success
    vpReturn = (void*)(spCell + 1);
  }
  else
  {
    // record the error
    ulErrReport(vpCtx, APG_MEM_ID, __LINE__, __FILE__);
    vpReturn = NULL;
  }

  return vpReturn;
}

//////////////////////////////////////////////////////////////////////////
// DESCRIPTION: free previously allocated memory
//
// FUNCTION: void vMemFree
//
//      ARG: 1) void* vpCtx
//              : pointer to a memory context previously returned
//                from vpMemCtor() or vpMemCtorA()
//
//      ARG: 2) void* vpData
//              : pointer to the memory allocation to free
//                may be NULL, no error, no action
//
//   RETURN: void
//
//////////////////////////////////////////////////////////////////////////
void vMemFree(void* vpCtx, void* vpData)
{
  APG_MEM_CTX*  spCtx = (APG_MEM_CTX*)vpCtx;
  APG_MEM_CELL* spCell;
  DASSERT(spCtx && spCtx->vpValidate == (void*)spCtx);
  if(vpData)
  {
    spCell = (APG_MEM_CELL*)vpData;
    --spCell;
    
    // pop from active list
    APG_MEMSTATS(&spCtx->sStats, MEM_ID_FREE, spCell, 0);
    vActivePop(spCtx, spCell);
  }
}

//////////////////////////////////////////////////////////////////////////
// DESCRIPTION: re-allocates memory (upsized or downsized)
//              - any previous data on the memory is copied to the new memory location
//
// FUNCTION: void* vpMemRealloc
//
//      ARG: 1) void* vpCtx
//              : pointer to a memory context previously returned
//                from vpMemCtor() or vpMemCtorA()
//
//      ARG: 2) void* vpData
//              : pointer to the memory allocation to free
//                may be NULL, no error, no action
//
//      ARG: 3) ulong ulBytes
//              : new number of bytes of memory to allocate
//                if 0, no action, no error
//
//   RETURN: returns a pointer to the re-allocated memory
//            NULL if no memory is allocated
//
//////////////////////////////////////////////////////////////////////////
void* vpMemRealloc(void* vpCtx, void* vpData, ulong ulBytes)
{
  void*         vpReturn = NULL;
  APG_MEM_CTX*  spCtx = (APG_MEM_CTX*)vpCtx;
  APG_MEM_CELL* spOldCell;
  APG_MEM_CELL* spNewCell;
  void*         vpDst;
  void*         vpSrc;
  ulong         ulCopy;
  DASSERT(spCtx && spCtx->vpValidate == (void*)spCtx);

  if(vpData && ulBytes)
  {
    spOldCell = (APG_MEM_CELL*)vpData;
    --spOldCell;
    
    // get the new allocation
    spNewCell = (APG_MEM_CELL*)spCtx->pfnAllocator(ulBytes + sizeof(APG_MEM_CELL));
    if(spNewCell)
    {
      spNewCell->spNext = spOldCell->spNext;
      spNewCell->spPrev = spOldCell->spPrev;
      spOldCell->spNext->spPrev = (struct APG_MEM_CELL_STRUCT*)spNewCell;
      spOldCell->spPrev->spNext = (struct APG_MEM_CELL_STRUCT*)spNewCell;
      spNewCell->ulSeq = spOldCell->ulSeq;
      spNewCell->ulSize = ulBytes;
      
      // copy the data from old allocation to new
      ulCopy = min(ulBytes, spOldCell->ulSize);
      vpDst = (void*)(spNewCell + 1);
      vpSrc = (void*)(spOldCell + 1);
      memcpy(vpDst, vpSrc, ulCopy);
      
      // free the old data
      APG_MEMSTATS(&spCtx->sStats, MEM_ID_REALLOC, spOldCell, spNewCell);
      spCtx->pfnDeAllocator((void*)spOldCell);
      vpReturn = (void*)(spNewCell + 1);
    }
    else
    {
      // record the error
      ulErrReport(vpCtx, APG_MEM_ID, __LINE__, __FILE__);
      vpReturn = NULL;
    }
  }

  return vpReturn;
}

//////////////////////////////////////////////////////////////////////////
// DESCRIPTION: get a check point to the current state of the memory component
// This can be used to restore the memory allocations to the current state.
//
// FUNCTION: ulong ulMemCheckPoint
//
//      ARG: 1) void* vpCtx
//              : pointer to a memory context previously returned
//                from vpMemCtor() or vpMemCtorA()
//
//   RETURN: returns a check point index
//
//////////////////////////////////////////////////////////////////////////
ulong ulMemCheckPoint(void* vpCtx)
{
  APG_MEM_CTX*  spCtx = (APG_MEM_CTX*)vpCtx;
  APG_MEM_CELL* spLast;
  ulong         ulChk;
  DASSERT(spCtx && spCtx->vpValidate == (void*)spCtx);

  if(spCtx->spActiveList)
  {
    spLast = (APG_MEM_CELL*)spCtx->spActiveList->spPrev;
    ulChk = spLast->ulSeq + 1;
  }
  else
  {
    ulChk = 0;
  }

  return ulChk;
}

//////////////////////////////////////////////////////////////////////////
// DESCRIPTION: restore the memory component to a previous state
//
// FUNCTION: void vMemFreeToCheckPoint
//
//      ARG: 1) void* vpCtx
//              : pointer to a memory context previously returned
//                from vpMemCtor() or vpMemCtorA()
//
//      ARG: 2) ulong ulChk
//              : previous state index
//                must be a value returned from a previous call to ulMemCheckPoint()
//
//   RETURN: void
//
//////////////////////////////////////////////////////////////////////////
void vMemFreeToCheckPoint(void* vpCtx, ulong ulChk)
{
  APG_MEM_CTX*  spCtx = (APG_MEM_CTX*)vpCtx;
  APG_MEM_CELL* spLast;
  DASSERT(spCtx && spCtx->vpValidate == (void*)spCtx);

  // release all active blocks above the check point
  while(APG_TRUE)
  {
    // break if list is empty
    if(spCtx->spActiveList == NULL){break;}
    
    // remove the last cell from the active list
    spLast = (APG_MEM_CELL*)spCtx->spActiveList->spPrev;

    // break if this cell is below check point
    if(spLast->ulSeq < ulChk){break;}

    // remove the cell from the active list
    APG_MEMSTATS(&spCtx->sStats, MEM_ID_FREE, spLast, 0);
    vActivePop(spCtx, spLast);
  }
}

//////////////////////////////////////////////////////////////////////////
// DESCRIPTION: release all memory currently allocated.
//  restore the memory component to the initial state
//
// FUNCTION: void vMemClear
//
//      ARG: 1) void* vpCtx
//              : pointer to a memory context previously returned
//                from vpMemCtor() or vpMemCtorA()
//
//   RETURN: void
//
//////////////////////////////////////////////////////////////////////////
void vMemClear(void* vpCtx)
{
  vMemFreeToCheckPoint(vpCtx, 0);
}

//////////////////////////////////////////////////////////////////////////
// STATIC HELPER FUNCTIONS
//////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////
// DESCRIPTION: Updates memory statistics.
//
// FUNCTION: static void vMemStats
//
//      ARG: 1) APG_MEM_STATS* spStats
//              : pointer to a statistics struct
//
//      ARG: 2) ulong ulID
//              : MEM_ID_ALLOC - update allocation stats on allocation
//              : MEM_ID_REALLOC - update reallocation stats
//              : MEM_ID_FREE - update allocation stats on deallocation
//
//      ARG: 3) APG_MEM_CELL* spIn
//              : pointer to the memory cell being re/de/allocated
//
//      ARG: 4) APG_MEM_CELL* spOut
//              : pointer to the new, reallocated memory cell
//
//   RETURN: void
//
//////////////////////////////////////////////////////////////////////////
static void vMemStats(APG_MEM_STATS* spStats, ulong ulID, APG_MEM_CELL* spIn, APG_MEM_CELL* spOut)
{
  switch(ulID)
  {
  case MEM_ID_ALLOC:
    ++spStats->ulAllocations;
    ++spStats->ulCells;
    spStats->ulHeapBytes += spIn->ulSize + sizeof(APG_MEM_CELL);
    if(spStats->ulCells > spStats->ulMaxCells)
    {
      spStats->ulMaxCells = spStats->ulCells;
    }
    if(spStats->ulHeapBytes > spStats->ulMaxHeapBytes)
    {
      spStats->ulMaxHeapBytes = spStats->ulHeapBytes;
    }
    break;
  case MEM_ID_REALLOC:
    ++spStats->ulReAllocations;
    spStats->ulHeapBytes += spOut->ulSize - spIn->ulSize;
    if(spStats->ulHeapBytes > spStats->ulMaxHeapBytes)
    {
      spStats->ulMaxHeapBytes = spStats->ulHeapBytes;
    }
    break;
  case MEM_ID_FREE:
    ++spStats->ulFrees;
    --spStats->ulCells;
    spStats->ulHeapBytes -= spIn->ulSize + sizeof(APG_MEM_CELL);
    break;
  }
  
}

//////////////////////////////////////////////////////////////////////////
// DESCRIPTION: push a new memory cell on the linked list
//
// FUNCTION: static void vActivePush
//
//      ARG: 1) APG_MEM_CTX* spCtx
//              : pointer to a memory context previously returned
//                from vpMemCtor() or vpMemCtorA()
//
//      ARG: 2) APG_MEM_CELL* spCellIn
//              : the new cell to be pushed
//
//   RETURN: void
//
//////////////////////////////////////////////////////////////////////////
static void vActivePush(APG_MEM_CTX* spCtx, APG_MEM_CELL* spCellIn)
{
  struct APG_MEM_CELL_STRUCT* spLast;
  struct APG_MEM_CELL_STRUCT* spFirst;
  struct APG_MEM_CELL_STRUCT* spCell = (struct APG_MEM_CELL_STRUCT*)spCellIn;

  // sequence number roll over is a fatal error
  // link the cell
  switch(spCtx->ulActiveCellCount)
  {
  case 0:
    spCtx->spActiveList = spCell;
    spCell->spNext = spCell->spPrev = spCell;
    spCell->ulSeq = 0;
    break;
  case 1:
    spLast = spCtx->spActiveList;
    spLast->spNext = spLast->spPrev = spCell;
    spCell->spNext = spCell->spPrev = spLast;
    spCell->ulSeq = spLast->ulSeq + 1;
    break;
  default:
    spFirst = spCtx->spActiveList;
    spLast = spFirst->spPrev;
    spFirst->spPrev = spCell;
    spLast->spNext = spCell;
    spCell->spNext = spFirst;
    spCell->spPrev = spLast;
    spCell->ulSeq = spLast->ulSeq + 1;
    break;
  }

  ++spCtx->ulActiveCellCount;
}

//////////////////////////////////////////////////////////////////////////
// DESCRIPTION: removes (pops) a memory cell from the linked list
//
// FUNCTION: static void vActivePop
//
//      ARG: 1) APG_MEM_CTX* spCtx
//              : pointer to a memory context previously returned
//                from vpMemCtor() or vpMemCtorA()
//
//      ARG: 2) APG_MEM_CELL* spCellIn
//              : the memory cell to be removed (popped)
//
//   RETURN: void
//
//////////////////////////////////////////////////////////////////////////
static void vActivePop(APG_MEM_CTX* spCtx, APG_MEM_CELL* spCellIn)
{
  struct APG_MEM_CELL_STRUCT* spPrev;
  struct APG_MEM_CELL_STRUCT* spNext;
  struct APG_MEM_CELL_STRUCT* spCell = (struct APG_MEM_CELL_STRUCT*)spCellIn;

  if(spCtx->spActiveList == spCell)
  {
    // update the pointer to the first cell
    if(spCtx->ulActiveCellCount == 1)
    {
      spCtx->spActiveList = NULL;
    }
    else
    {
      spCtx->spActiveList = spCtx->spActiveList->spNext;
    }
  }

  // remove the cell from the active list
  spPrev = spCell->spPrev;
  spNext = spCell->spNext;
  spPrev->spNext = spCell->spNext;
  spNext->spPrev = spCell->spPrev;
  
  --spCtx->ulActiveCellCount;

  // free the data
  spCtx->pfnDeAllocator((void*)spCell);
}
