/********************************************************************
  APG - an ABNF Parser Generator
  Copyright  (C) 2007 Coast to Coast Research, Inc.
 
  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 2 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program.  If not, see
  <http://www.gnu.org/licenses/old-licenses/gpl-2.0.html>
  or write to the Free Software Foundation, Inc.,
  51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

    author: Lowell Thomas
            lowell@coasttocoastresearch.com
            http://www.coasttocoastresearch.com

   created: 8/25/2007 10:45
  filename: D:\Projects\Apgv5_0\ApgUtilities\Display.c
   purpose: Various useful display functions.
  
*********************************************************************/
#include "Apg.h"
#include "Private.h"
#include "Operators.h"
#include "OperatorsPPT.h"
#include "ApgUtilities.h"

// display alert info and terminate
//////////////////////////////////////////////////////////////////////////
// DESCRIPTION: Displays the location and source identifier of an alert.
//              then terminates the process.
//
// FUNCTION: void vTerminalAlert
//
//      ARG: 1) int iID
//              : alert identifier
//
//      ARG: 2) unsigned int uiLine
//              : line number of the alert (__LINE__)
//
//      ARG: 3) char* cpFile
//              : file name where the alert occurred (__FILE__)
//
//   RETURN: none - process is terminated with the alert identifier.
//
//////////////////////////////////////////////////////////////////////////
void vTerminalAlert(int iID, unsigned int uiLine, char* cpFile)
{
  vPrintAlertState(iID, uiLine, cpFile);
  printf("Terminating with exit code %d\n", iID);
  exit(iID);
}

// print alert information
//////////////////////////////////////////////////////////////////////////
// DESCRIPTION: Simply prints the alert state and returns.
//
//      ARG: 1) int iID
//              : alert identifier
//
//      ARG: 2) unsigned int uiLine
//              : line number of the alert (__LINE__)
//
//      ARG: 3) char* cpFile
//              : file name where the alert occurred (__FILE__)
//
//   RETURN: void
//
//////////////////////////////////////////////////////////////////////////
void vPrintAlertState(int iID, unsigned int uiLine, char* cpFile)
{
  printf("\n");
  printf("Alert State:\n");
  printf("ID   = %d\n", iID);
  printf("file = %s\n", cpFile);
  printf("line = %d\n", (int)uiLine);
}

//////////////////////////////////////////////////////////////////////////
// DESCRIPTION: prints a (one-based) numbered list of the lines in an ASCII file
// file must be ASCII null-term
// line numbers are one-based
// line ender is CRLF, CR or LF
// last line may or may not have line ender
//
// FUNCTION: void vDisplaySourceLines
//
//      ARG: 1) char* cpSrc
//              : pointer to the file of ASCII-text lines
//
//      ARG: 2) FILE* spOut
//              : pointer to a valid, open I/O device as returned from fopen()
//
//   RETURN: void
//
//////////////////////////////////////////////////////////////////////////
void   vDisplaySourceLines(char* cpSrc, FILE* spOut)
{
  ulong ulCharBeg = 0;
  ulong ulCharEnd = 0;
  ulong ulLine = 1;
  char* cpLine = cpSrc;
  char* cpBeg = cpSrc;
  char* cpEnd = cpBeg + strlen(cpSrc) + 1;
  DASSERT(cpSrc && spOut);

  printf(" line  char\n");
  for(; cpBeg < cpEnd; ++cpBeg, ++ulCharEnd)
  {
    if(*cpBeg == 0)
    {
      // print the last line in case it has no line ender
      fprintf(spOut, "[%04lu, %04lu]: %s\n", ulLine, ulCharBeg, cpLine);
    }
    
    if(*cpBeg == 13)
    {
      char cSave = *cpBeg;
      if(cpBeg[1] == 10)
      {
        // line ends with CRLF
        *cpBeg = 0;
        fprintf(spOut, "[%04lu, %04lu]: %s<CRLF>\n", ulLine, ulCharBeg, cpLine);
        *cpBeg = cSave;
        ulLine++;
        ulCharBeg += ulCharEnd + 2;
        ulCharEnd = -1;
        cpBeg += 1;
        cpLine = cpBeg + 1;
        continue;
      }
      else
      {
        // line ends with CR only
        *cpBeg = 0;
        fprintf(spOut, "[%04lu, %04lu]: %sCR\n", ulLine, ulCharBeg, cpLine);
        *cpBeg = cSave;
        ulLine++;
        ulCharBeg += ulCharEnd + 1;
        ulCharEnd = -1;
        cpLine = cpBeg + 1;
        continue;
      }
    }
    
    if(*cpBeg == 10)
    {
      char cSave = *cpBeg;

      // line ends with LR only
      *cpBeg = 0;
      fprintf(spOut, "[%04lu, %04lu]: %sLF\n", ulLine, ulCharBeg, cpLine);
      *cpBeg = cSave;
      ulLine++;
      ulCharBeg += ulCharEnd + 1;
      ulCharEnd = -1;
      cpLine = cpBeg + 1;
      continue;
    }
  }
}

//////////////////////////////////////////////////////////////////////////
// DESCRIPTION: print a parser state
//
// FUNCTION: void vPrintState
//
//      ARG: 1) const APG_PARSER_STATE* spState
//              : pointer to an APG parser state struct
//
//      ARG: 2) char* cpTitle
//              : pointer to a null-termed title, may be NULL
//
//   RETURN: 
//
//////////////////////////////////////////////////////////////////////////
void vPrintState(const APG_PARSER_STATE* spState, char* cpTitle)
{
  char* cpState;
  DASSERT(spState);

  switch(spState->ulState)
  {
  case SYN_PRE:
    cpState = "ACTIVE";
    break;
  case SYN_NOMATCH:
    cpState = "NOMATCH";
    break;
  case SYN_EMPTY:
    cpState = "EMPTY";
    break;
  case SYN_MATCH:
    cpState = "MATCH";
    break;
  default:
    cpState = "UNKNOWN";
    break;
  }

  printf("\n");
  if(cpTitle) printf("%s\n", cpTitle);
  printf("*** Parser State ***\n");
  printf("  state = %s\n", cpState);
  printf(" length = %lu\n", spState->ulStringLength);
  printf("matched = %lu\n", spState->ulCharactersMatched);
}


//////////////////////////////////////////////////////////////////////////
// DESCRIPTION: prints statistics for one opcode type
//
// FUNCTION: static void vPrintOperatorLine
//
//      ARG: 1) char* cpOp
//              : pointer to the name of the opcode type
//
//      ARG: 2) APG_NODE_STATS* spNode
//              : pointer to the statistics struct
//
//   RETURN: void
//
//////////////////////////////////////////////////////////////////////////
static void vPrintOperatorLine(char* cpOp, APG_NODE_STATS* spNode)
{
  DASSERT(cpOp);
  printf("%5s %7lu %7lu %7lu %7lu",
    cpOp,
    spNode->ulVisits,
    spNode->ulMatched,
    spNode->ulEmpty,
    spNode->ulNotMatched);
}

//////////////////////////////////////////////////////////////////////////
// DESCRIPTION: Print parser statistics
//
// FUNCTION: void vPrintStats
//
//      ARG: 1) void* vpCtx
//              : pointer to a valid parser component context
//
//   RETURN: void
//
//////////////////////////////////////////////////////////////////////////
void vPrintStats(void* vpCtx)
{
  APG_PARSER_CTX* spCtx = (APG_PARSER_CTX*)vpCtx;
  DASSERT(spCtx && spCtx->vpValidate == (void*)spCtx);

  printf("\n");
  printf("*************************\n");
  printf("*** Parser Statistics ***\n");
  printf("*************************\n");
  printf("input string length     = %lu\n", spCtx->sSrc.ulLen);
  printf("max characters matched  = %lu\n", spCtx->sStats.ulCharsMaxMatch);
  printf("maximum tree depth      = %lu\n", spCtx->sStats.ulNodesMaxDepth);
  printf("back track occurrences  = %lu\n", spCtx->sStats.ulBackTrackCount);
  printf("back tracked characters = %lu\n", spCtx->sStats.ulBackTrackChars);

  printf("\n");
//  printf("OPERATOR VISITS\n");
  printf("%5s %7s %7s %7s %7s %7s %7s\n", "", "VISITS", "MATCH", "EMPTY", "NOMATCH", "ACTIVE", "CHILDREN");
  vPrintOperatorLine("RNM", &spCtx->sStats.sRnm);
  printf("\n");
  vPrintOperatorLine("ALT", &spCtx->sStats.sAlt);
  printf("          %7lu\n", spCtx->sStats.sAlt.ulChildCount);
  vPrintOperatorLine("CAT", &spCtx->sStats.sCat);
  printf("          %7lu\n", spCtx->sStats.sCat.ulChildCount);
  vPrintOperatorLine("REP", &spCtx->sStats.sRep);
  printf("          %7lu\n", spCtx->sStats.sRep.ulChildCount);
  vPrintOperatorLine("PRD", &spCtx->sStats.sPrd);
  printf("\n");
  vPrintOperatorLine("TRG", &spCtx->sStats.sTrg);
  printf("\n");
  vPrintOperatorLine("TBS", &spCtx->sStats.sTbs);
  printf("\n");
  vPrintOperatorLine("TLS", &spCtx->sStats.sTls);
  printf("\n");
  vPrintOperatorLine("PPT", &spCtx->sStats.sPpt);
  printf(" %7lu\n", spCtx->sStats.sPpt.ulPossibleMatched);
  vPrintOperatorLine("EOF", &spCtx->sStats.sEof);
  printf("\n");
  vPrintOperatorLine("TOTAL", &spCtx->sStats.sTotal);
  printf(" %7lu  %7lu\n", spCtx->sStats.sTotal.ulPossibleMatched, spCtx->sStats.sTotal.ulChildCount);
}

//////////////////////////////////////////////////////////////////////////
// DESCRIPTION: Counts the total number of node visits for all opcode types.
//
// FUNCTION: ulong ulNodeVisits
//
//      ARG: 1) void* vpCtx
//              : pointer to a valid parser component context
//
//   RETURN: void
//
//////////////////////////////////////////////////////////////////////////
ulong ulNodeVisits(void* vpCtx)
{
  ulong ulReturn = 0;
  APG_PARSER_CTX* spCtx = (APG_PARSER_CTX*)vpCtx;
  DASSERT(spCtx && spCtx->vpValidate == (void*)spCtx);

  ulReturn += spCtx->sStats.sAlt.ulVisits;
  ulReturn += spCtx->sStats.sCat.ulVisits;
  ulReturn += spCtx->sStats.sRnm.ulVisits;
  ulReturn += spCtx->sStats.sRep.ulVisits;
  ulReturn += spCtx->sStats.sPrd.ulVisits;
  ulReturn += spCtx->sStats.sTrg.ulVisits;
  ulReturn += spCtx->sStats.sTbs.ulVisits;
  ulReturn += spCtx->sStats.sTls.ulVisits;

  return ulReturn;
}

//////////////////////////////////////////////////////////////////////////
// DESCRIPTION: Prints statistics gathered by a memory component
//
// FUNCTION: static ulong ulMemPrintStats
//
//      ARG: 1) void* vpCtx
//              : pointer to a valid memory component context
//
//   RETURN: true on success
//
//////////////////////////////////////////////////////////////////////////
static ulong   ulMemPrintStats(void* vpCtx)
{
  APG_MEM_STATS* spStats;
  APG_MEM_CTX* spCtx = (APG_MEM_CTX*)vpCtx;
  DASSERT(spCtx && spCtx->vpValidate == (void*)spCtx);

  spStats = &spCtx->sStats;
  printf("Allocations     = %lu\n", spStats->ulAllocations);
  printf("Re-Allocations  = %lu\n", spStats->ulReAllocations);
  printf("Frees           = %lu\n", spStats->ulFrees);
  printf("Cells           = %lu\n", spStats->ulCells);
  printf("Heap Bytes      = %lu\n", spStats->ulHeapBytes);
  printf("Max Cells       = %lu\n", spStats->ulMaxCells);
  printf("Max Heap Bytes  = %lu\n", spStats->ulMaxHeapBytes);
  
  return APG_TRUE;
}
//////////////////////////////////////////////////////////////////////////
// DESCRIPTION: Prints statistics gathered by a memory component
//
// FUNCTION: void vPrintMemStats
//
//      ARG: 1) void* vpCtx
//              : pointer to a valid memory component context
//
//   RETURN: void
//
//////////////////////////////////////////////////////////////////////////
void vPrintMemStats(void* vpCtx)
{
  ulong ulStats;
  APG_MEM_CTX* spCtx = (APG_MEM_CTX*)vpCtx;
  DASSERT(spCtx && spCtx->vpValidate == (void*)spCtx);

  printf("\n");
  printf("*************************\n");
  printf("*** MEMORY STATISTICS ***\n");
  printf("*************************\n");
  ulStats = ulMemPrintStats(vpCtx);
  if(!ulStats)
  {
    printf("<none>\n");
  }
}

//////////////////////////////////////////////////////////////////////////
// DESCRIPTION: Prints a number of blanks to indent a line before it is printed
//
// FUNCTION: void vIndent
//
//      ARG: 1) ulong ulIndent
//              : number of blanks to print for indenting
//
//   RETURN: void
//
//////////////////////////////////////////////////////////////////////////
void vIndent(ulong ulIndent)
{
  while(ulIndent--)
  {
    printf(" ");
  }
}
//////////////////////////////////////////////////////////////////////////
// DESCRIPTION: Print a visual display of a AST, indicating traversal
//              direction, using indentations to indicate tree depth.
//
// FUNCTION: void vDisplayParseTree
//
//      ARG: 1) void* vpCtx
//              : pointer to a valid parser component context
//
//   RETURN: void
//
//////////////////////////////////////////////////////////////////////////
void vDisplayParseTree(void* vpCtx)
{
  APG_RULES*      spRuleList;
  APG_PT_NODE*    spNext;
  APG_PT_NODE*    spEnd;
  APG_PT_NODE*    spPtNodes;
  APG_PT_VALUE*   spPtValues;
  APG_PT_VALUE*   spValue;
  PFN_CALLBACK    pfnCallBack;
  ulong           ulValue, ulValueLen;
  ulong           ulIndent = 0;
  char            caBuf[1024];
  APG_PARSER_CTX* spCtx = (APG_PARSER_CTX*)vpCtx;
  DASSERT(spCtx && spCtx->vpValidate == (void*)spCtx);
  
  spRuleList = (APG_RULES*)spCtx->vpRules;
  if(spCtx->vpPtNodes)
  {
    spPtNodes = (APG_PT_NODE*)vpVecFront(spCtx->vpPtNodes);
    spPtValues = (APG_PT_VALUE*)vpVecFront(spCtx->vpPtValues);
    spNext = spPtNodes;
    spEnd = spNext + ulVecSize(spCtx->vpPtNodes);
    for(; spNext < spEnd; ++spNext)
    {
      DASSERT((spNext->ulPairedNode != APG_PT_UNDEFINED));
      spValue = spPtValues + spNext->ulValue;
      ulValue = spValue->sValue.ulOffset;
      ulValueLen = spValue->sValue.ulLen;
      pfnCallBack = spRuleList[spValue->ulRuleIndex].pfnCallBack;
      if(pfnCallBack)
      {
        // process this node
        if(spNext->ulDirection == APG_PT_DOWN)
        {
          // indent
          vIndent(ulIndent);

          // print the rule name
          memcpy((void*)caBuf, (void*)(spCtx->sSrc.ucpStr + ulValue), ulValueLen);
          caBuf[ulValueLen] = 0;
          printf("[%s][%lu, %lu] %s\n",
            spRuleList[spValue->ulRuleIndex].cpRuleName,
            ulValue, ulValueLen,
            caBuf);

          // increment indent
          ++ulIndent;
        }
        else
        {
          // decrement indent
          --ulIndent;

          // print the rule name
          vIndent(ulIndent);
          printf("[%s][UP]\n", spRuleList[spValue->ulRuleIndex].cpRuleName);
        }
      }
    }
  }
}

//////////////////////////////////////////////////////////////////////////
// DESCRIPTION: Print a human-readable version of a PPT.
//
// FUNCTION: void vPrintPPT
//
//      ARG: 1) uchar* ucpPPT
//              : pinter to the PPT to display
//
//      ARG: 2) ulong ulCharSetSize
//              : the number of characters in the PPT
//
//   RETURN: 
//
//////////////////////////////////////////////////////////////////////////
void vPrintPPT(uchar* ucpPPT, ulong ulCharSetSize)
{
  ulong ulChar;
  ulong ulLineChar;
  ulong ulLineLen;
  ulong ulValue;
  ulong ulLineBits;
  ulong ulLineValues;
  char  caBits[128];
  char  caValues[128];
  ulong ul1_Bit;
  ulong ulEOF = ulCharSetSize - 1;
  DASSERT(ucpPPT && ulCharSetSize > 0);
  
  if(PPT_IS_1BIT(ucpPPT))
  {
    ul1_Bit = APG_TRUE;
    printf(" 1-bit PPT\n");
  } 
  else
  {
    ul1_Bit = APG_FALSE;
    printf(" 2-bit PPT\n");
  }
  
  ulLineLen = 20;
  ulLineChar = 0;
  ulLineBits = 0;
  ulLineValues = 0;
  for(ulChar = 0; ulChar < ulCharSetSize; ++ulChar)
  {
    ulValue = ulGetPPTValue(ucpPPT, ulChar);
    if(ulChar == ulEOF)
    {
      ulLineBits += sprintf(&caBits[ulLineBits], " EOF");
    } 
    else
    {
      ulLineBits += sprintf(&caBits[ulLineBits], " %03lu", ulChar);
    }
    if(ul1_Bit)
    {
      switch(ulValue)
      {
      case EMPTY:
        ulLineValues += sprintf(&caValues[ulLineValues], " ERR");
        break;
      case MATCH:
        ulLineValues += sprintf(&caValues[ulLineValues], " ERR");
        break;
      case ACTIVE:
        ulLineValues += sprintf(&caValues[ulLineValues], "   A");
        break;
      case NOMATCH:
        ulLineValues += sprintf(&caValues[ulLineValues], "   N");
        break;
      default:
        ulLineValues += sprintf(&caValues[ulLineValues], " ERR");
        break;
      }
    } 
    else
    {
      switch(ulValue)
      {
      case EMPTY:
        ulLineValues += sprintf(&caValues[ulLineValues], "   E");
        break;
      case MATCH:
        ulLineValues += sprintf(&caValues[ulLineValues], "   M");
        break;
      case ACTIVE:
        ulLineValues += sprintf(&caValues[ulLineValues], "   A");
        break;
      case NOMATCH:
        ulLineValues += sprintf(&caValues[ulLineValues], "   N");
        break;
      default:
        ulLineValues += sprintf(&caValues[ulLineValues], " ERR");
        break;
      }
    }
    
    ++ulLineChar;
    if(ulLineChar >= ulLineLen)
    {
      ulLineChar = 0;
      ulLineBits = 0;
      ulLineValues = 0;
      printf("%s\n", caBits);
      printf("%s\n", caValues);
    }
  }
  
  if(ulLineBits > 0)
  {
    printf("%s\n", caBits);
    printf("%s\n", caValues);
  }
}

//////////////////////////////////////////////////////////////////////////
// DESCRIPTION: Prints all of the low-level error messages associated with
//              a memory component.
//
// FUNCTION: void vPrintErrors
//
//      ARG: 1) void* vpCtx
//              : pointer to a valid memory component context.
//
//   RETURN: void
//
//////////////////////////////////////////////////////////////////////////
void vPrintErrors(void* vpCtx)
{
  ulong     ulErrors, ulIndex, ulTest;
  APG_ERR sError;
  char*     cpId;
  APG_MEM_CTX* spMemCtx = (APG_MEM_CTX*)vpCtx;
  APG_ERR_CTX* spCtx = spMemCtx->spErrCtx;
  DASSERT(spMemCtx->vpValidate == (void*)spMemCtx);

  ulErrors = ulErrCount(vpCtx);
  printf("\n");
  printf("***\n");
  printf("*** ERROR MESSAGES[%lu]\n", ulErrors);
  printf("***\n");
  if(ulErrors)
  {
    for(ulIndex = 0; ulIndex < ulErrors; ++ulIndex)
    {
      ulTest = ulErrMsg(vpCtx, ulIndex, &sError);
      if(ulTest)
      {
        int iId = (int)sError.ulID;
        switch(iId)
        {
        case APG_DBG_ID:
          cpId = "debug error";
          break;
        case APG_MEM_ID:
          cpId = "memory component error (out of memory)";
          break;
        case APG_VEC_ID:
          cpId = "vector component error";
        	break;
        case APG_PAR_ID:
          cpId = "parser component error";
          break;
        case APG_CAT_ID:
          cpId = "catalog component error";
          break;
        default:
          cpId = NULL;
            break;
        }
        
        if(cpId)
        {
          printf("ID   = %d, (%s)\n", iId, cpId);
        }
        else
        {
          printf("ID   = %d\n", iId);
        }
        printf("line = %lu\n", sError.ulLine);
        printf("file = %s\n", sError.cpFile);
      }
      else
      {
        printf("<none>\n");
      }
      printf("\n");
    }
  }
  else
  {
    printf("<none>\n");
  }
}

//////////////////////////////////////////////////////////////////////////
// DESCRIPTION: Converts carriage return and line feed characters to
//              a printable ASCII character for display.
//
// FUNCTION: static vConvertLineEnd
//
//      ARG: 1) char* cpBuf
//              : 
//
//      ARG: 2) ulong ulBufLen
//              : 
//
//   RETURN: 
//
//////////////////////////////////////////////////////////////////////////
static vConvertLineEnd(char* cpBuf, ulong ulBufLen)
{
  ulong             ul;

  if(cpBuf && ulBufLen)
  {
    for(ul = 0; ul < ulBufLen; ++ul)
    {
      if(cpBuf[ul] == 10 || cpBuf[ul] == 13)
      {
        // convert CR & LF to ~
        cpBuf[ul] = '~';
      }
    }
  }
}

// print some statistics
//////////////////////////////////////////////////////////////////////////
// DESCRIPTION: Prints Packrat statistics.
//
// FUNCTION: void vPrintPackratRuleStats
//
//      ARG: 1) void* vpCtx
//              : pointer to a valid Packrat statistics component context
//
//   RETURN: void
//
//////////////////////////////////////////////////////////////////////////
void vPrintPackratRuleStats(void* vpCtx)
{
  ulong       ulRule;
  ulong       ulChar;
  ulong       ulEntries;
  ulong       ulNoHits;
  ulong       ulSingleHits;
  ulong       ulMultipleHits;
  APG_RULES* spRuleList;
  APG_RULES* spRule;
  APG_PACKRAT* spPackrat;
  APG_PACKRAT_CTX* spPackratCtx;
  APG_PARSER_CTX* spCtx = (APG_PARSER_CTX*)vpCtx;
  DASSERT(spCtx && spCtx->vpValidate == (void*)spCtx);
  if(!spCtx->vpPackrat)
  {
    goto APG_EXIT;
  }

  spPackratCtx = (APG_PACKRAT_CTX*)spCtx->vpPackrat;
  DASSERT(spPackratCtx && spPackratCtx->vpValidate == (void*)spPackratCtx);
  
  // allocate and initialize the packrat table
  spRuleList = (APG_RULES*)spCtx->vpRules;
  
  // rules: hit counts
  ulEntries = 0;
  ulNoHits = 0;
  ulSingleHits = 0;
  ulMultipleHits = 0;
  spRule = spRuleList;
  spPackrat = spPackratCtx->spTable;
  for(ulRule = 0; ulRule < spPackratCtx->ulRuleCount; ++ulRule, ++spRule, spPackrat += spPackratCtx->ulCharCount)
  {
    for(ulChar = 0; ulChar < spPackratCtx->ulCharCount; ++ulChar, spPackrat)
    {
      ++ulEntries;
      if(spPackrat[ulChar].ulCount == 0){++ulNoHits;}
      if(spPackrat[ulChar].ulCount == 1){++ulSingleHits;}
      if(spPackrat[ulChar].ulCount > 1){++ulMultipleHits;}
    }
  }
  
  printf("\n");
  printf("*** Packrat Rule Stats\n");
  printf("*** Packrat hit counts\n");
  printf("%4lu rules * chars\n", ulEntries);
  printf("%4lu no hits\n", ulNoHits);
  printf("%4lu single hits\n", ulSingleHits);
  printf("%4lu multiple hits\n", ulMultipleHits);
  
  
  //rules: single hits
//   printf("\n");
//   printf("*** Packrat single hits\n");
//   spRule = spRuleList;
//   spPackrat = spPackratCtx->spTable;
//   for(ulRule = 0; ulRule < spPackratCtx->ulRuleCount; ++ulRule, ++spRule, spPackrat += spPackratCtx->ulCharCount)
//   {
//     for(ulChar = 0; ulChar < spPackratCtx->ulCharCount; ++ulChar, spPackrat)
//     {
//       ++ulEntries;
//       if(spPackrat[ulChar].ulCount == 1)
//       {
//         printf("[%10s, %4lu] 1 hit\n", spRule->cpRuleName, ulChar);
//       }
//     }
//   }
  
  // rules: multiple hits
  printf("\n");
  printf("*** Packrat multiple hits\n");
  spRule = spRuleList;
  spRule = spRuleList;
  spRule = spRuleList;
  spPackrat = spPackratCtx->spTable;
  for(ulRule = 0; ulRule < spPackratCtx->ulRuleCount; ++ulRule, ++spRule, spPackrat += spPackratCtx->ulCharCount)
  {
    for(ulChar = 0; ulChar < spPackratCtx->ulCharCount; ++ulChar, spPackrat)
    {
      ++ulEntries;
      if(spPackrat[ulChar].ulCount > 1)
      {
        printf("[%10s, %4lu] %lu hits\n", spRule->cpRuleName, ulChar, spPackrat[ulChar].ulCount);
      }
    }
  }

APG_EXIT:
  return;
}

//////////////////////////////////////////////////////////////////////////
// DESCRIPTION: Prints the copyright and no-warranty notices.
//
// FUNCTION: void vLicenseNotice
//
//   RETURN: void
//
//////////////////////////////////////////////////////////////////////////
void vLicenseNotice()
{
  static char* cpNotice =
    "APG - an ABNF Parser Generator\n"
    "Copyright (C) 2007 Coast to Coast Research, Inc.\n"
    "This program comes with ABSOLUTELY NO WARRANTY. This is free software,\n"
    "and you are welcome to redistribute it under certain conditions.\n"
    "For warranty and conditions see the GNU General Public License version 2\n"
    "included with this distribution or at\n"
    "<http://www.gnu.org/licenses/old-licenses/gpl-2.0.html>\n"
    "or write to the Free Software Foundation, Inc.,\n"
    "51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.\n";
  printf("%s\n\n", cpNotice);
}
