/********************************************************************
  APG - an ABNF Parser Generator
  Copyright  (C) 2007 Coast to Coast Research, Inc.
 
  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 2 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program.  If not, see
  <http://www.gnu.org/licenses/old-licenses/gpl-2.0.html>
  or write to the Free Software Foundation, Inc.,
  51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

    author: Lowell Thomas
            lowell@coasttocoastresearch.com
            http://www.coasttocoastresearch.com

   created: 8/25/2007 12:54
  filename: D:\Projects\Apgv5_0\Generator\Errors.c
   purpose: Error and warning reporting component. Used to report
            syntax errors in the ABNF grammar.
  
*********************************************************************/
#include "Generator.h"

typedef struct
{
  void* vpValidate; // used to validate the context
  void* vpLinesCtx; // context of the grammar lines catalog
  void* vpErrCtx;   // context of the erratum struct
  void* vpMsgCtx;   // context of the erratum text messages
  ulong ulIterIndex;// index of errata message of next iteration
} GRAMERR_CTX;

typedef struct
{
  ulong ulLineIndex;
  ulong ulMsgIndex;
} GRAMERR;

//////////////////////////////////////////////////////////////////////////
// DESCRIPTION: Error component constructor
//
// FUNCTION: void* vpGrammarErrorsCtor
//
//      ARG: 1) void* vpMemCtx
//              : pointer to a valid memory component context
//
//      ARG: 2) void* vpLinesCtx
//              : pointer to a vector of pointers, one to each line 
//                in the grammar file.
//
//   RETURN: pointer to the GrammarError component context
//
//////////////////////////////////////////////////////////////////////////
void* vpGrammarErrorsCtor(void* vpMemCtx, void* vpLinesCtx)
{
  void* vpReturn = NULL;
  GRAMERR_CTX* spCtx;
  DASSERT(vpMemCtx && vpLinesCtx);

  // allocate & clear the data
  spCtx = (GRAMERR_CTX*)vpMemAlloc(vpMemCtx, sizeof(GRAMERR_CTX));
  UASSERT(spCtx, vpMemCtx, APG_GEN_ID);
  memset((void*)spCtx, 0, sizeof(GRAMERR_CTX));
  spCtx->vpLinesCtx = vpLinesCtx;

  // allocate a vector of erratum instances
  spCtx->vpErrCtx = vpVecCtor(vpMemCtx, sizeof(GRAMERR), 0);
  UASSERT(spCtx->vpErrCtx, vpMemCtx, APG_GEN_ID);

  // allocate a vector of erratum messages
  spCtx->vpMsgCtx = vpVecCtor(vpMemCtx, sizeof(char), 0);
  UASSERT(spCtx->vpMsgCtx, vpMemCtx, APG_GEN_ID);

  // success
  spCtx->vpValidate = (void*)spCtx;
  vpReturn = (void*)spCtx;

APG_EXIT:
  if(!vpReturn)
  {
    if(spCtx)
    {
      // release all memory allocated here
      spCtx->vpValidate = NULL;
      vMemDtor(vpMemCtx);
    }
  }
  return vpReturn;
}

//////////////////////////////////////////////////////////////////////////
// DESCRIPTION: Error component destructor
//
// FUNCTION: void vGrammarErrorsDtor
//
//      ARG: 1) void* vpErrorsCtx
//              : pointer to a valid GrammarErrors component 
//
//   RETURN: void
//
//////////////////////////////////////////////////////////////////////////
void  vGrammarErrorsDtor(void* vpErrorsCtx)
{
  GRAMERR_CTX* spCtx = (GRAMERR_CTX*)vpErrorsCtx;
  DASSERT(spCtx && spCtx->vpValidate == (void*)spCtx);

  // release all memory allocations
  if(spCtx->vpErrCtx)vVecDtor(spCtx->vpErrCtx);
  if(spCtx->vpMsgCtx)vVecDtor(spCtx->vpMsgCtx);
  spCtx->vpValidate = NULL;
}

//////////////////////////////////////////////////////////////////////////
// DESCRIPTION: Report an error. Pushes and error report on the vector
//              of error messages
//
// FUNCTION: ulong ulGrammarErrorsReport
//
//      ARG: 1) void* vpErrorsCtx
//              : pointer to a valid GrammarErrors component 
//
//      ARG: 2) ulong ulLineIndex
//              : index of the line where the error occurred
//
//      ARG: 3) char* cpMsg
//              : message to associate with the error
//
//   RETURN: true on success
//
//////////////////////////////////////////////////////////////////////////
ulong ulGrammarErrorsReport(void* vpErrorsCtx, ulong ulLineIndex, char* cpMsg)
{
  ulong ulReturn = APG_FALSE;
  ulong ulMsgIndex;
  ulong ulLen;
  void* vpTest;
  char  cNull = 0;
  GRAMERR* spErr;
  GRAMERR_CTX* spCtx = (GRAMERR_CTX*)vpErrorsCtx;
  DASSERT(spCtx && spCtx->vpValidate == (void*)spCtx);

  ulLen = (ulong)strlen(cpMsg);
  if(ulLen >= APG_ERROR_BUFLEN)
  {
    // truncate the message
    ulLen = APG_ERROR_BUFLEN - 1;
  }

  // push the message on the msg vector
  ulMsgIndex = ulVecSize(spCtx->vpMsgCtx);
  vpTest = vpVecPushn(spCtx->vpMsgCtx, (void*)cpMsg, ulLen);
  if(!vpTest){goto APG_EXIT;}
  vpTest = vpVecPush(spCtx->vpMsgCtx, (void*)&cNull);
  if(!vpTest){goto APG_EXIT;}

  // push the errata on the err vector
  spErr = (GRAMERR*)vpVecPush(spCtx->vpErrCtx, NULL);
  if(!spErr){goto APG_EXIT;}
  spErr->ulLineIndex = ulLineIndex;
  spErr->ulMsgIndex = ulMsgIndex;
  
  // success
  ulReturn = APG_TRUE;

APG_EXIT:
  return ulReturn;
}

//////////////////////////////////////////////////////////////////////////
// DESCRIPTION: Returns the number of errors on the error queue.
//
// FUNCTION: ulong ulGrammarErrorsCount
//
//      ARG: 1) void* vpErrorsCtx
//              : pointer to a valid GrammarErrors component 
//
//   RETURN: Returns the number of errors on the error queue.
//
//////////////////////////////////////////////////////////////////////////
ulong ulGrammarErrorsCount(void* vpErrorsCtx)
{
  GRAMERR_CTX* spCtx = (GRAMERR_CTX*)vpErrorsCtx;
  DASSERT(spCtx && spCtx->vpValidate == (void*)spCtx);
  
  return ulVecSize(spCtx->vpErrCtx);
}

//////////////////////////////////////////////////////////////////////////
// DESCRIPTION: Initializes an iteration over the error queue.
//
// FUNCTION: void vGrammarErrorsInit
//
//      ARG: 1) void* vpErrorsCtx
//              : pointer to a valid GrammarErrors component 
//
//   RETURN: void
//
//////////////////////////////////////////////////////////////////////////
void  vGrammarErrorsInit(void* vpErrorsCtx)
{
  GRAMERR_CTX* spCtx = (GRAMERR_CTX*)vpErrorsCtx;
  DASSERT(spCtx && spCtx->vpValidate == (void*)spCtx);
  spCtx->ulIterIndex = 0;
}

//////////////////////////////////////////////////////////////////////////
// DESCRIPTION: Returns the error information for the next error in the queue
//
// FUNCTION: ulong vGrammarErrorsNext
//
//      ARG: 1) void* vpErrorsCtx
//              : pointer to a valid GrammarErrors component 
//
//      ARG: 2) ulong* ulpSrcLineNo
//              : grammar source line number of erratum returned here
//
//      ARG: 3) char* cpSrcLine
//              : pointer to a buffer to receive the null-term text of the grammar source line in error
//          NOTE: cpSrcLine/cpErrMsg buffers must be at least APG_ERROR_BUFLEN in length
//
//      ARG: 4) char* cpErrMsg
//              : pointer to a buffer to receive the null-term text errata message
//
//   RETURN: true on success
//
//////////////////////////////////////////////////////////////////////////
ulong vGrammarErrorsNext(void* vpErrorsCtx, ulong* ulpSrcLineNo, char* cpSrcLine, char* cpErrMsg)
{
  ulong ulReturn = APG_FALSE;
  GRAMERR* spErrata;
  APG_BSTR* spSrc;
  char* cpErr;
  GRAMERR_CTX* spCtx = (GRAMERR_CTX*)vpErrorsCtx;
  DASSERT(spCtx && spCtx->vpValidate == (void*)spCtx);

  // initialize
  *ulpSrcLineNo = UNDEFINED;
  *cpSrcLine = 0;
  *cpErrMsg = 0;

  // get the next errata
  spErrata = (GRAMERR*)vpVecAt(spCtx->vpErrCtx, spCtx->ulIterIndex);
  if(!spErrata){goto APG_EXIT;}

  // get the errata
  spSrc = (APG_BSTR*)vpVecAt(spCtx->vpLinesCtx, spErrata->ulLineIndex);
  if(!spSrc){goto APG_EXIT;}

  // get the errata message
  cpErr = (char*)vpVecAt(spCtx->vpMsgCtx, spErrata->ulMsgIndex);
  if(!cpErr){goto APG_EXIT;}

  // success
  *ulpSrcLineNo = spErrata->ulLineIndex + 1;
  ulCatalogGetString(spSrc, cpSrcLine);
  strcpy(cpErrMsg, cpErr);
  ++spCtx->ulIterIndex;
  ulReturn = APG_TRUE;

APG_EXIT:
  return ulReturn;
}
