/********************************************************************
  APG - an ABNF Parser Generator
  Copyright  (C) 2007 Coast to Coast Research, Inc.
 
  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 2 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program.  If not, see
  <http://www.gnu.org/licenses/old-licenses/gpl-2.0.html>
  or write to the Free Software Foundation, Inc.,
  51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

    author: Lowell Thomas
            lowell@coasttocoastresearch.com
            http://www.coasttocoastresearch.com

   created: 8/25/2007 12:57
  filename: D:\Projects\Apgv5_0\Generator\NonPPTOpcodes.c
   purpose: Converts intermediate opcodes to production format.
  
*********************************************************************/
#include "Generator.h"

// static functions
static ulong ulCountOpcodeBytes(GENCTX* spData, OP_INT_ALL* spOps, ulong ulFrom, ulong ulTo);
static ulong ulProductionRules(GENCTX* spData, INT_RULE_LIST* spIntRules, ulong ulRuleCount, void* vpRulesCtx);

// returns opcode count, zero on error
//////////////////////////////////////////////////////////////////////////
// DESCRIPTION: Process a list of intermediate opcodes into a list of
//              production opcodes usable by a parser component.
//
// FUNCTION: ulong ulProductionOpcodes
//
//      ARG: 1) GENCTX* spData
//              : pointer to a valid generator component context
//
//   RETURN: true on success
//
//////////////////////////////////////////////////////////////////////////
ulong ulProductionOpcodes(GENCTX* spData)
{
  ulong ulReturn = APG_FALSE;
  OP_ALT  sAlt;
  OP_CAT  sCat;
  OP_RNM  sRnm;
  OP_REP  sRep;
  OP_PRD  sPrd;
  OP_TRG  sTrg;
  OP_TBS  sTbs;
  OP_TLS  sTls;
  APG_RULES*      spRules;
  OP_INT_ALL*     spOp;
  OP_INT_ALL*     spOpEnd;
  ulong           ulOpcodeBytes;
  OP_INT_ALL*     spOpBeg;
  INT_RULE_LIST*  spIntRule;
  APG_RULES*      spProdRule;
  APG_RULES*      spProdRuleEnd;
  ulong           ulThisOffset;
  void*           vpTest;
  ulong           ulTest;

    // create a production rule list
    spData->vpRuleCtx = vpVecCtor(spData->vpMemCtx, sizeof(APG_RULES), spData->ulRuleCount);
    GENASSERT(spData->vpRuleCtx);

    spIntRule = (INT_RULE_LIST*)vpVecFront(spData->vpIntRuleCtx);
    ulTest = ulProductionRules(spData, spIntRule, spData->ulRuleCount, spData->vpRuleCtx);
    GENASSERT(ulTest);

    // create a vector for the production opcodes
    spData->vpOpcodeCtx = vpVecCtor(spData->vpMemCtx, 1, sizeof(OP_TLS) * ulVecSize(spData->vpIntOpcodesCtx));
    GENASSERT(spData->vpOpcodeCtx);

    // initialize the loop over all opcodes
    spRules = (APG_RULES*)vpVecFront(spData->vpRuleCtx);
    spOpBeg = (OP_INT_ALL*)vpVecFront(spData->vpIntOpcodesCtx);
    spProdRule = spRules;
    spProdRuleEnd = spRules + spData->ulRuleCount;
    spOp = spOpBeg;
    spData->ulOpcodeCount = 0;
    for(; spProdRule < spProdRuleEnd; ++spProdRule, ++spIntRule)
    {
      DASSERT(spOp == &spOpBeg[spIntRule->ulOpcodeOffset]);
      ulThisOffset = ulVecSize(spData->vpOpcodeCtx);
      spProdRule->vpOp = (void*)(ulThisOffset);// index here - converted to pointer in ulParserInit()
      spOpEnd = spOp + spIntRule->ulOpcodeCount;
      for(; spOp < spOpEnd; ++spOp)
      {
        ++spData->ulOpcodeCount;
        ulThisOffset = ulVecSize(spData->vpOpcodeCtx);
        ulOpcodeBytes = ulCountOpcodeBytes(spData, spOpBeg, spOp->ulIndexThis, spOp->ulIndexLast);
        switch(spOp->ulOpId)
        {
        case OP_INT_ID_ALT:
          sAlt.pfnOp = (PFN_OP)OP_ID_ALT;
          sAlt.spNext = (OP_GEN*)(ulThisOffset + ulOpcodeBytes);
          vpTest = vpVecPushn(spData->vpOpcodeCtx, (void*)&sAlt, sizeof(OP_ALT));
          GENASSERT(vpTest);
          break;
        case OP_INT_ID_CAT:
          sCat.pfnOp = (PFN_OP)OP_ID_CAT;
          sCat.spNext = (OP_GEN*)(ulThisOffset + ulOpcodeBytes);
          vpTest = vpVecPushn(spData->vpOpcodeCtx, (void*)&sCat, sizeof(OP_CAT));
          GENASSERT(vpTest);
          break;
        case OP_INT_ID_REP:
          sRep.pfnOp = (PFN_OP)OP_ID_REP;
          sRep.spNext = (OP_GEN*)(ulThisOffset + ulOpcodeBytes);
          sRep.ulMin = spOp->ulMin;
          sRep.ulMax = spOp->ulMax;
          vpTest = vpVecPushn(spData->vpOpcodeCtx, (void*)&sRep, sizeof(OP_REP));
          GENASSERT(vpTest);
          break;
        case OP_INT_ID_RNM:
          sRnm.pfnOp = (PFN_OP)OP_ID_RNM;
          sRnm.spNext = (OP_GEN*)(ulThisOffset + ulOpcodeBytes);
          sRnm.ulRuleIndex = spOp->ulRuleIndex;
          vpTest = vpVecPushn(spData->vpOpcodeCtx, (void*)&sRnm, sizeof(OP_RNM));
          GENASSERT(vpTest);
          break;
        case OP_INT_ID_TRG:
          sTrg.pfnOp = (PFN_OP)OP_ID_TRG;
          sTrg.spNext = (OP_GEN*)(ulThisOffset + ulOpcodeBytes);
          sTrg.ulMin = spOp->ulMin;
          sTrg.ulMax = spOp->ulMax;
          vpTest = vpVecPushn(spData->vpOpcodeCtx, (void*)&sTrg, sizeof(OP_TRG));
          GENASSERT(vpTest);
          break;
        case OP_INT_ID_TBS:
          sTbs.pfnOp = (PFN_OP)OP_ID_TBS;
          sTbs.spNext = (OP_GEN*)(ulThisOffset + ulOpcodeBytes);
          sTbs.ucpString = (uchar*)spOp->ulOffset;
          sTbs.ulLength = spOp->ulLength;
          vpTest = vpVecPushn(spData->vpOpcodeCtx, (void*)&sTbs, sizeof(OP_TBS));
          GENASSERT(vpTest);
          break;
        case OP_INT_ID_TLS:
          sTls.pfnOp = (PFN_OP)OP_ID_TLS;
          sTls.spNext = (OP_GEN*)(ulThisOffset + ulOpcodeBytes);
          sTls.ucpString = (uchar*)spOp->ulOffset;
          sTls.ulLength = spOp->ulLength;
          vpTest = vpVecPushn(spData->vpOpcodeCtx, (void*)&sTls, sizeof(OP_TLS));
          GENASSERT(vpTest);
          break;
        case OP_INT_ID_PRD:
          sPrd.pfnOp = (PFN_OP)OP_ID_PRD;
          sPrd.spNext = (OP_GEN*)(ulThisOffset + ulOpcodeBytes);
          sPrd.ulAnd = spOp->ulAnd;
          vpTest = vpVecPushn(spData->vpOpcodeCtx, (void*)&sPrd, sizeof(OP_PRD));
          GENASSERT(vpTest);
          break;
        default:
          GENASSERT(APG_FALSE);
          break;
        }
      }
    }
    
    // remove unneeded intermediate rules
    vVecDtor(spData->vpIntRuleCtx);
    spData->vpIntRuleCtx = NULL;
    
    // remove unneeded intermediate opcodes
    vVecDtor(spData->vpIntOpcodesCtx);
    spData->vpIntOpcodesCtx = NULL;
    
    if(spData->sConfig.ulDisplayOpcodes)
    {
      vPrintProductionOpcodes(spData);
    }
    // success
    ulReturn = APG_TRUE;

APG_EXIT:
  return ulReturn;
}

//////////////////////////////////////////////////////////////////////////
// DESCRIPTION: Count the bytes in a list of production opcode.
//
// FUNCTION: static ulong ulCountOpcodeBytes
//
//      ARG: 1) GENCTX* spData
//              : pointer to a valid generator component context
//
//      ARG: 2) OP_INT_ALL* spOps
//              : pointer to a list of intermediate opcode
//
//      ARG: 3) ulong ulFrom
//              : first opcode in the list to count
//
//      ARG: 4) ulong ulTo
//              : last opcode in the list to count
//
//   RETURN: the number of bytes in the designated opcode list
//
//////////////////////////////////////////////////////////////////////////
static ulong ulCountOpcodeBytes(GENCTX* spData, OP_INT_ALL* spOps, ulong ulFrom, ulong ulTo)
{
  ulong ulReturn = 0;
  ulong ulBytes = 0;
  for(; ulFrom <= ulTo; ++ulFrom)
  {
    switch(spOps[ulFrom].ulOpId)
    {
    case OP_INT_ID_ALT:
      ulBytes += sizeof(OP_ALT);
      break;
    case OP_INT_ID_CAT:
      ulBytes += sizeof(OP_CAT);
      break;
    case OP_INT_ID_REP:
      ulBytes += sizeof(OP_REP);
      break;
    case OP_INT_ID_RNM:
      ulBytes += sizeof(OP_RNM);
      break;
    case OP_INT_ID_TRG:
      ulBytes += sizeof(OP_TRG);
      break;
    case OP_INT_ID_TBS:
      ulBytes += sizeof(OP_TBS);
      break;
    case OP_INT_ID_TLS:
      ulBytes += sizeof(OP_TLS);
      break;
    case OP_INT_ID_PRD:
      ulBytes += sizeof(OP_PRD);
      break;
    default:
      GENASSERT(APG_FALSE);
      break;
    }
  }
  
  ulReturn = ulBytes;

APG_EXIT:
  return ulBytes;
}

//////////////////////////////////////////////////////////////////////////
// DESCRIPTION: Translates an intermediate rule list to a production rule list.
//
// FUNCTION: static ulong ulProductionRules
//
//      ARG: 1) GENCTX* spData
//              : pointer to a valid generator component context
//
//      ARG: 2) INT_RULE_LIST* spIntRules
//              : intemediate rule list
//
//      ARG: 3) ulong ulRuleCount
//              : number of rules in the list
//
//      ARG: 4) void* vpRulesCtx
//              : vector context for the production rule list.
//
//   RETURN: true on success
//
//////////////////////////////////////////////////////////////////////////
static ulong ulProductionRules(GENCTX* spData, INT_RULE_LIST* spIntRules, ulong ulRuleCount, void* vpRulesCtx)
{
  ulong     ulReturn = APG_FALSE;
  ulong     ul;
  APG_RULES sProdRule;
  void*     vpTest;
  
  for(ul = 0; ul < ulRuleCount; ++ul)
  {
    INT_RULE_LIST*  spIntRule = &spIntRules[ul];
    sProdRule.cpRuleName = (char*)spIntRule->ulNameOffset;  // index here - converted to pointer in ulParserInit()
    sProdRule.usSyntax = spIntRule->usSyntax;
    sProdRule.usSemantics = spIntRule->usSemantics;
    sProdRule.pfnCallBack = (PFN_CALLBACK)ul;               // index here - converted to pointer in ulParserInit()
    sProdRule.vpOp = NULL;                                  // computed in ulProductionOpcodes()
    vpTest = vpVecPush(vpRulesCtx, (void*)&sProdRule);
    GENASSERT(vpTest);
  }
  
  ulReturn = APG_TRUE;

APG_EXIT:
  return ulReturn;
}
