/********************************************************************
  APG - an ABNF Parser Generator
  Copyright  (C) 2007 Coast to Coast Research, Inc.
 
  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 2 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program.  If not, see
  <http://www.gnu.org/licenses/old-licenses/gpl-2.0.html>
  or write to the Free Software Foundation, Inc.,
  51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

    author: Lowell Thomas
            lowell@coasttocoastresearch.com
            http://www.coasttocoastresearch.com

   created: 8/25/2007 13:04
  filename: D:\Projects\Apgv5_0\Generator\SESTWalker.c
   purpose: Generic depth-first traversal of the SEST. Calls user-defined
            call back functions to perform specialized translations
            of the SEST.
  
*********************************************************************/
#include "Generator.h"

// SESTWalker provides the shell logic for doing a depth-first traversal of a Single Expansion Syntax Tree
// The Single Expansion Syntax Tree at REP and PRD nodes differs between the recursion analysis traversal
// and the PPT computation traversal. All other nodes have the same logic for both.
// Therefore, SESTWalker provides different REP and PRD nodes. At construction time, the user
// must specify which type of traversal is to be done.

typedef void (*PFN_NODEFUNC)(void* vpCtx, OP_INT_ALL* spOp);

typedef struct 
{
  void*           vpMemCtx;
  void*           vpErrorsCtx;   // grammar errors reporting component
  uchar*          ucpStrTbl;
  OP_INT_ALL*     spIntOpcodes;
  INT_RULE_LIST*  spIntRules;
  void*           vpCBData;
  PFN_SESTCB      pfnaCallBacks[OP_INT_ID_MAX];
  ulong           ulRuleCount;
  ulong           ulTreeDepth;
  ulong           ulMaxTreeDepth;
  ulong           ulTrace;
  ulong           ulMaxTerminals;
  ulong           ulTerminals;

  // type of traversal
  PFN_NODEFUNC pfnRnm;
  PFN_NODEFUNC pfnAlt;
  PFN_NODEFUNC pfnCat;
  PFN_NODEFUNC pfnPrd;
  PFN_NODEFUNC pfnRep;
  PFN_NODEFUNC pfnTrg;
  PFN_NODEFUNC pfnTbs;
  PFN_NODEFUNC pfnTls;
} SESTWALKERCTX;

// execute each opcode
static ulong ulExecuteOp(void* vpCtx, OP_INT_ALL* spOp);

// Single Expansion Syntax Tree walker operator functions
static void vSESTRNM(void* vpCtx, OP_INT_ALL* spOp);
static void vSESTALT(void* vpCtx, OP_INT_ALL* spOp);
static void vSESTCAT(void* vpCtx, OP_INT_ALL* spOp);
static void vSESTREP(void* vpCtx, OP_INT_ALL* spOp);
static void vPPTREP(void* vpCtx, OP_INT_ALL* spOp);
static void vSESTPRD(void* vpCtx, OP_INT_ALL* spOp);
static void vSESTTRG(void* vpCtx, OP_INT_ALL* spOp);
static void vSESTTBS(void* vpCtx, OP_INT_ALL* spOp);
static void vSESTTLS(void* vpCtx, OP_INT_ALL* spOp);

// default call back functions
static ulong ulDefaultRNM(ulong ulState, ulong ulTreeDepth, ulong ulChildCount, ulong ulCountMax, OP_INT_ALL* spOp, void* vpData);
static ulong ulDefaultALT(ulong ulState, ulong ulTreeDepth, ulong ulChildCount, ulong ulCountMax, OP_INT_ALL* spOp, void* vpData);
static ulong ulDefaultCAT(ulong ulState, ulong ulTreeDepth, ulong ulChildCount, ulong ulCountMax, OP_INT_ALL* spOp, void* vpData);
static ulong ulDefaultREP(ulong ulState, ulong ulTreeDepth, ulong ulChildCount, ulong ulCountMax, OP_INT_ALL* spOp, void* vpData);
static ulong ulDefaultPRD(ulong ulState, ulong ulTreeDepth, ulong ulChildCount, ulong ulCountMax, OP_INT_ALL* spOp, void* vpData);
static ulong ulDefaultTRG(ulong ulState, ulong ulTreeDepth, ulong ulChildCount, ulong ulCountMax, OP_INT_ALL* spOp, void* vpData);
static ulong ulDefaultTBS(ulong ulState, ulong ulTreeDepth, ulong ulChildCount, ulong ulCountMax, OP_INT_ALL* spOp, void* vpData);
static ulong ulDefaultTLS(ulong ulState, ulong ulTreeDepth, ulong ulChildCount, ulong ulCountMax, OP_INT_ALL* spOp, void* vpData);

// tracing functions
#define _TRACE_DOWN 0
#define _TRACE_UP   1
void vTrace(SESTWALKERCTX* spCtx, OP_INT_ALL* spOp, ulong ulUp);

//////////////////////////////////////////////////////////////////////////
// GENERIC TRAVERSAL OF THE Single Expansion Syntax Tree OF INTERMEDIATE OPCODES
//////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////
// DESCRIPTION: SESTWalker component constructor
//
// FUNCTION: void* vpSESTCtor
//
//      ARG: 1) GENCTX* spData
//              : pointer to a valid generator component context
//
//      ARG: 2) ulong ulTraversal
//              : SEST_RECURSION for a recursion analysis traversal
//              : SEST_PPT for a PPT analysis traversal
//
//      ARG: 3) void* vpCBData
//              : user-defined data to be passed to the call back functions.
//
//   RETURN: on success, a pointer to a valid SESTWalker component context
//           on failure, NULL
//
//////////////////////////////////////////////////////////////////////////
void* vpSESTCtor(GENCTX* spData, ulong ulTraversal, void* vpCBData)
{
  // allocate a context
  SESTWALKERCTX* spCtx = (SESTWALKERCTX*)vpMemAlloc(spData->vpMemCtx, sizeof(SESTWALKERCTX));
  if(spCtx)
  {
    // initialize the context
    memset((void*)spCtx, 0, sizeof(SESTWALKERCTX));
    spCtx->vpMemCtx = spData->vpMemCtx;
    spCtx->vpErrorsCtx = spData->vpErrorsCtx;
    spCtx->ucpStrTbl = (uchar*)vpVecFront(spData->vpStrTblCtx);
    spCtx->spIntOpcodes = (OP_INT_ALL*)vpVecFront(spData->vpIntOpcodesCtx);
    spCtx->spIntRules = (INT_RULE_LIST*)vpVecFront(spData->vpIntRuleCtx);
    spCtx->ulRuleCount = spData->ulRuleCount;
    spCtx->vpCBData = vpCBData;
    spCtx->ulTreeDepth = 0;

    // traversal type
    spCtx->pfnRnm = vSESTRNM;
    spCtx->pfnAlt = vSESTALT;
    spCtx->pfnCat = vSESTCAT;
    spCtx->pfnPrd = vSESTPRD;
    spCtx->pfnTrg = vSESTTRG;
    spCtx->pfnTbs = vSESTTBS;
    spCtx->pfnTls = vSESTTLS;
    switch(ulTraversal)
    {
    case SEST_RECURSION:
      spCtx->pfnRep = vSESTREP;
    	break;
    case SEST_PPT:
      spCtx->pfnRep = vPPTREP;
        break;
    default:
      GENASSERT(APG_FALSE);
        break;
    }

    // set the default call back functions
    spCtx->pfnaCallBacks[OP_INT_ID_RNM] = ulDefaultRNM;
    spCtx->pfnaCallBacks[OP_INT_ID_ALT] = ulDefaultALT;
    spCtx->pfnaCallBacks[OP_INT_ID_CAT] = ulDefaultCAT;
    spCtx->pfnaCallBacks[OP_INT_ID_REP] = ulDefaultREP;
    spCtx->pfnaCallBacks[OP_INT_ID_PRD] = ulDefaultPRD;
    spCtx->pfnaCallBacks[OP_INT_ID_TRG] = ulDefaultTRG;
    spCtx->pfnaCallBacks[OP_INT_ID_TBS] = ulDefaultTBS;
    spCtx->pfnaCallBacks[OP_INT_ID_TLS] = ulDefaultTLS;
  }

APG_EXIT:
  return (void*)spCtx;
}

//////////////////////////////////////////////////////////////////////////
// DESCRIPTION: SESTWalker component destructor
//
// FUNCTION: void vSESTDtor
//
//      ARG: 1) void* vpCtx
//              : pointer to a valid SESTWalker component context
//
//   RETURN: void
//
//////////////////////////////////////////////////////////////////////////
void  vSESTDtor(void* vpCtx)
{
  SESTWALKERCTX* spCtx = (SESTWALKERCTX*)vpCtx;
  if(spCtx)
  {
    vMemFree(spCtx->vpMemCtx, vpCtx);
  }
}

//////////////////////////////////////////////////////////////////////////
// DESCRIPTION: set the call back functions to user-defined functions
//
// FUNCTION: void vSESTSetCallBacks
//
//      ARG: 1) void* vpCtx
//              : pointer to a valid SESTWalker component context
//
//      ARG: 2) PFN_SESTCB* ppfnCB
//              : array of at least OP_INT_ID_MAX call back function pointers
//                individual pointers in the list NULL
//
//   RETURN: void
//
//////////////////////////////////////////////////////////////////////////
void vSESTSetCallBacks(void* vpCtx, PFN_SESTCB* ppfnCB)
{
  SESTWALKERCTX* spCtx = (SESTWALKERCTX*)vpCtx;
  ulong ulIndex;
  for(ulIndex = 0; ulIndex < OP_INT_ID_MAX; ++ulIndex)
  {
    spCtx->pfnaCallBacks[ulIndex] = ppfnCB[ulIndex];
  }
}

// walk (depth-first traversal) of the specified rule's Single Expansion Syntax Tree
//////////////////////////////////////////////////////////////////////////
// DESCRIPTION: walk (depth-first traversal) of the specified rule's
//              Single Expansion Syntax Tree
//
// FUNCTION: ulong ulSESTWalk
//
//      ARG: 1) void* vpCtx
//              : pointer to a valid SESTWalker component context
//
//      ARG: 2) ulong ulStartRule
//              : rule index of the start rule
//
//      ARG: 3) ulong ulMaxTerminals
//              : maximum number of SEST nodes to process before giving up
//
//      ARG: 4) ulong ulTrace
//              : true for a trace of the traversal, false otherwise
//
//   RETURN: true on success
//           false if ulMaxTerminals has been exceeded
//
//////////////////////////////////////////////////////////////////////////
ulong ulSESTWalk(void* vpCtx, ulong ulStartRule, ulong ulMaxTerminals, ulong ulTrace)
{
  SESTWALKERCTX*  spCtx = (SESTWALKERCTX*)vpCtx;
  OP_INT_ALL      sRnm;
  INT_RULE_LIST*  spRule;
  ulong           ulTreeDepth;
  ulong           ulReturn;
  ulong           ulRule;
  PFN_SESTCB      pfnCB = spCtx->pfnaCallBacks[OP_INT_ID_RNM];

  // set the trace parameter
  spCtx->ulTrace = ulTrace;
  spCtx->ulMaxTerminals = ulMaxTerminals;
  spCtx->ulTreeDepth = 0;
  ulTreeDepth = 0;
  spCtx->ulMaxTreeDepth = 0;
  spCtx->ulTerminals = 0;

  // initialize all rules closed
  for(ulRule = 0; ulRule < spCtx->ulRuleCount; ++ulRule)
  {
    spCtx->spIntRules[ulRule].ulLeaf = APG_FALSE;
  }

  // get the rule
  spRule = &spCtx->spIntRules[ulStartRule];

  // create a dummy RNM for the
  memset((void*)&sRnm, 0, sizeof(sRnm));
  sRnm.ulOpId = OP_INT_ID_RNM;
  sRnm.ulRuleIndex = ulStartRule;
  sRnm.ulIndexThis = spRule->ulOpcodeOffset;
  sRnm.ulIndexLast = spRule->ulOpcodeOffset + spRule->ulOpcodeCount - 1;

  // walk the tree
  ulReturn = ulExecuteOp(vpCtx, &sRnm);

  // validate the completion of the walk
  DASSERT(spCtx->ulTreeDepth == ulTreeDepth);

  return ulReturn;
}

//////////////////////////////////////////////////////////////////////////
// DESCRIPTION: Returns the max SEST depth after a traversal.
//
// FUNCTION: ulong ulSESTMaxTreeDepth
//
//      ARG: 1) void* vpCtx
//              : pointer to a valid SESTWalker component context
//
//   RETURN: Returns the max SEST depth after a traversal.
//
//////////////////////////////////////////////////////////////////////////
ulong ulSESTMaxTreeDepth(void* vpCtx)
{
  SESTWALKERCTX*  spCtx = (SESTWALKERCTX*)vpCtx;
  return spCtx->ulMaxTreeDepth;
}

//////////////////////////////////////////////////////////////////////////
// STATIC HELPER FUNCTIONS
//////////////////////////////////////////////////////////////////////////
// execute the tree node operation
//////////////////////////////////////////////////////////////////////////
// DESCRIPTION: execute a single tree node operation
//
// FUNCTION: static ulong ulExecuteOp
//
//      ARG: 1) void* vpCtx
//              : pointer to a valid SESTWalker component context
//
//      ARG: 2) OP_INT_ALL* spOp
//              : the opcode for this node
//
//   RETURN: true on success
//           false if ulMaxTerminals has been exceeded
//
//////////////////////////////////////////////////////////////////////////
static ulong ulExecuteOp(void* vpCtx, OP_INT_ALL* spOp)
{
  SESTWALKERCTX*  spCtx = (SESTWALKERCTX*)vpCtx;
  ulong ulReturn = APG_TRUE;
  if(spCtx->ulTrace)
  {
    vTrace(spCtx, spOp, _TRACE_DOWN);
  }
  
  if(spCtx->ulTerminals < spCtx->ulMaxTerminals)
  {
    ++spCtx->ulTreeDepth;
    if(spCtx->ulTreeDepth > spCtx->ulMaxTreeDepth)
    {
      spCtx->ulMaxTreeDepth = spCtx->ulTreeDepth;
    }
    switch(spOp->ulOpId)
    {
    case OP_INT_ID_RNM:
      spCtx->pfnRnm(vpCtx, spOp);
      break;
    case OP_INT_ID_ALT:
      spCtx->pfnAlt(vpCtx, spOp);
      break;
    case OP_INT_ID_CAT:
      spCtx->pfnCat(vpCtx, spOp);
      break;
    case OP_INT_ID_REP:
      spCtx->pfnRep(vpCtx, spOp);
      break;
    case OP_INT_ID_PRD:
      spCtx->pfnPrd(vpCtx, spOp);
      break;
    case OP_INT_ID_TRG:
      spCtx->pfnTrg(vpCtx, spOp);
      break;
    case OP_INT_ID_TBS:
      spCtx->pfnTbs(vpCtx, spOp);
      break;
    case OP_INT_ID_TLS:
      spCtx->pfnTls(vpCtx, spOp);
      break;
    default:
      DASSERT(APG_FALSE);
      break;
    }
    
    --spCtx->ulTreeDepth;
    if(spCtx->ulTerminals >= spCtx->ulMaxTerminals)
    {
      // exit this node of the syntax tree
      ulReturn = APG_FALSE;
    }
  }
  else
  {
    ulReturn = APG_FALSE;
  }

  if(spCtx->ulTrace)
  {
    vTrace(spCtx, spOp, _TRACE_UP);
  }
  return ulReturn;
}

//////////////////////////////////////////////////////////////////////////
// Recursion analysis traversal node functions
//////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////
// DESCRIPTION: Recursion analysis traversal node functions
//
// FUNCTION: static void vSESTALT
//
//      ARG: 1) void* vpCtx
//              : pointer to a valid SESTWalker component context
//
//      ARG: 2) OP_INT_ALL* spOp
//              : pointer to the opcode for this node
//
//   RETURN: void
//
//////////////////////////////////////////////////////////////////////////
static void vSESTALT(void* vpCtx, OP_INT_ALL* spOp)
{
  SESTWALKERCTX*  spCtx = (SESTWALKERCTX*)vpCtx;
  OP_INT_ALL*     spOperand;
  ulong           ulReturn = PFN_OK;
  ulong           ulCount;
  ulong           ulContinue;
  PFN_SESTCB      pfnCB = spCtx->pfnaCallBacks[OP_INT_ID_ALT];

  // verify the operator type
  DASSERT(spOp->ulOpId == OP_INT_ID_ALT);

  // call all concatenations in this ALT
  ulCount = spOp->ulCount;
  spOperand = spOp + 1;
  if(pfnCB)
  {
    ulReturn = pfnCB(SEM_PRE, spCtx->ulTreeDepth, 0, spOp->ulCount, spOp, spCtx->vpCBData);
    if(ulReturn == PFN_OK)
    {
      // loop with call back functions
      for(ulCount = 1; ulCount <= spOp->ulCount; ++ulCount)
      {
        ulReturn = pfnCB(SEM_PRE, spCtx->ulTreeDepth, ulCount, spOp->ulCount, spOp, spCtx->vpCBData);
        if(ulReturn == PFN_SKIP){break;}
        ulContinue = ulExecuteOp(vpCtx, spOperand);
        if(!ulContinue){goto MAXTERM_EXIT;}
        ulReturn = pfnCB(SEM_POST, spCtx->ulTreeDepth, ulCount, spOp->ulCount, spOp, spCtx->vpCBData);
        if(ulReturn == PFN_SKIP){break;}
        spOperand += spOperand->ulIndexLast - spOperand->ulIndexThis + 1;
      }
      
      pfnCB(SEM_POST, spCtx->ulTreeDepth, 0, spOp->ulCount, spOp, spCtx->vpCBData);
    }
  }
  else
  {
    // loop without call back functions
    for(ulCount = 1; ulCount <= spOp->ulCount; ++ulCount)
    {
      ulContinue = ulExecuteOp(vpCtx, spOperand);
      if(!ulContinue){goto MAXTERM_EXIT;}
      spOperand += spOperand->ulIndexLast - spOperand->ulIndexThis + 1;
    }
  }
  
  // exit on reaching maximum terminal leaf node count
MAXTERM_EXIT:
  return;
}

static void vSESTCAT(void* vpCtx, OP_INT_ALL* spOp)
{
  SESTWALKERCTX*  spCtx = (SESTWALKERCTX*)vpCtx;
  OP_INT_ALL*     spOperand;
  ulong           ulReturn = PFN_OK;
  ulong           ulCount;
  ulong           ulContinue;
  PFN_SESTCB      pfnCB = spCtx->pfnaCallBacks[OP_INT_ID_CAT];
  
  // verify the operator type
  DASSERT(spOp->ulOpId == OP_INT_ID_CAT);
  
  // call all repetitions in this CAT
  ulCount = spOp->ulCount;
  spOperand = spOp + 1;
  if(pfnCB)
  {
    ulReturn = pfnCB(SEM_PRE, spCtx->ulTreeDepth, 0, spOp->ulCount, spOp, spCtx->vpCBData);
    if(ulReturn == PFN_OK)
    {
      // loop with call back functions
      for(ulCount = 1; ulCount <= spOp->ulCount; ++ulCount)
      {
        ulReturn = pfnCB(SEM_PRE, spCtx->ulTreeDepth, ulCount, spOp->ulCount, spOp, spCtx->vpCBData);
        if(ulReturn == PFN_SKIP){break;}
        ulContinue = ulExecuteOp(vpCtx, spOperand);
        if(!ulContinue){goto MAXTERM_EXIT;}
        ulReturn = pfnCB(SEM_POST, spCtx->ulTreeDepth, ulCount, spOp->ulCount, spOp, spCtx->vpCBData);
        if(ulReturn == PFN_SKIP){break;}
        spOperand += spOperand->ulIndexLast - spOperand->ulIndexThis + 1;
      }

      pfnCB(SEM_POST, spCtx->ulTreeDepth, 0, spOp->ulCount, spOp, spCtx->vpCBData);
    }
  }
  else
  {
    // loop without call back functions
    for(ulCount = 1; ulCount <= spOp->ulCount; ++ulCount)
    {
      ulContinue = ulExecuteOp(vpCtx, spOperand);
      if(!ulContinue){goto MAXTERM_EXIT;}
      spOperand += spOperand->ulIndexLast - spOperand->ulIndexThis + 1;
    }
  }
  
  // exit on reaching maximum terminal leaf node count
MAXTERM_EXIT:
  return;
}

static void vSESTREP(void* vpCtx, OP_INT_ALL* spOp)
{
  SESTWALKERCTX*  spCtx = (SESTWALKERCTX*)vpCtx;
  OP_INT_ALL*     spOperand;
  ulong           ulReturn = PFN_OK;
  ulong           ulContinue;
  PFN_SESTCB      pfnCB = spCtx->pfnaCallBacks[OP_INT_ID_REP];
  PFN_SESTCB      pfnAlt = spCtx->pfnaCallBacks[OP_INT_ID_ALT];
  OP_INT_ALL      sEmpty;
  OP_INT_ALL      sAlt;
  DASSERT(spOp->ulOpId == OP_INT_ID_REP);

  spOperand = spOp + 1;
  while(APG_TRUE)
  {
    if(pfnCB)
    {
      ulReturn = pfnCB(SEM_PRE, spCtx->ulTreeDepth, 0, 0, spOp, spCtx->vpCBData);
    }
    if(ulReturn == PFN_SKIP){break;}

    // execute the operand
    if(spOp->ulMin == 0 && spOp->ulMax == 0)
    {
      // just execute an empty terminal
      memset((void*)&sEmpty, 0, sizeof(sEmpty));
      sEmpty.ulOpId = OP_INT_ID_TLS;
      ulContinue = ulExecuteOp(spCtx, &sEmpty);
      if(!ulContinue){goto MAXTERM_EXIT;}
    }
    else if(spOp->ulMin == 0)
    {
      // mimic ALT with two children, operand / empty
      memset((void*)&sEmpty, 0, sizeof(sEmpty));
      sEmpty.ulOpId = OP_INT_ID_TLS;
      memset((void*)&sAlt, 0, sizeof(sAlt));
      sAlt.ulOpId = OP_INT_ID_ALT;
      sAlt.ulCount = 2;
      sAlt.ulIndexLast = 3;

      while(APG_TRUE)
      {
        if(pfnAlt)
        {
          ulReturn = pfnAlt(SEM_PRE, spCtx->ulTreeDepth, 0, 0, &sAlt, spCtx->vpCBData);
          if(ulReturn == PFN_SKIP){break;}
          ulReturn = pfnAlt(SEM_PRE, spCtx->ulTreeDepth, 1, 0, &sAlt, spCtx->vpCBData);
          if(ulReturn == PFN_SKIP){break;}
        }
        
        // execute the REP operand as the first ALT child
        ulContinue = ulExecuteOp(spCtx, spOperand);
        if(!ulContinue){goto MAXTERM_EXIT;}
        if(pfnAlt)
        {
          ulReturn = pfnAlt(SEM_POST, spCtx->ulTreeDepth, 1, 0, &sAlt, spCtx->vpCBData);
        }
        
        if(pfnAlt)
        {
          ulReturn = pfnAlt(SEM_PRE, spCtx->ulTreeDepth, 2, 0, &sAlt, spCtx->vpCBData);
        }
        if(ulReturn == PFN_SKIP){break;}
        
        // execute sEmpty, the second ALT child
        ulContinue = ulExecuteOp(spCtx, &sEmpty);
        if(!ulContinue){goto MAXTERM_EXIT;}
        if(pfnAlt)
        {
          ulReturn = pfnAlt(SEM_POST, spCtx->ulTreeDepth, 2, 0, &sAlt, spCtx->vpCBData);
          if(ulReturn == PFN_SKIP){break;}
          ulReturn = pfnAlt(SEM_POST, spCtx->ulTreeDepth, 0, 0, &sAlt, spCtx->vpCBData);
        }
        break;
      }
    }
    else
    {
      // just execute the single operand
      ulContinue = ulExecuteOp(spCtx, spOperand);
      if(!ulContinue){goto MAXTERM_EXIT;}
    }

    if(pfnCB)
    {
      pfnCB(SEM_POST, spCtx->ulTreeDepth, 0, 0, spOp, spCtx->vpCBData);
    }
    break;
  }
  
  // exit on reaching maximum terminal leaf node count
MAXTERM_EXIT:
  return;
}

static void vSESTPRD(void* vpCtx, OP_INT_ALL* spOp)
{
  SESTWALKERCTX*  spCtx = (SESTWALKERCTX*)vpCtx;
  ulong           ulReturn = PFN_OK;
  OP_INT_ALL*     spChild = spOp + 1;
  PFN_SESTCB      pfnCB = spCtx->pfnaCallBacks[OP_INT_ID_PRD];
  DASSERT(spOp->ulOpId == OP_INT_ID_PRD);
  
  if(pfnCB)
  {
    // execute call backs
    ulReturn = pfnCB(SEM_PRE, spCtx->ulTreeDepth, 0, 0, spOp, spCtx->vpCBData);
    if(ulReturn == PFN_OK)
    {
      ulExecuteOp(vpCtx, spChild);
      pfnCB(SEM_POST, spCtx->ulTreeDepth, 0, 0, spOp, spCtx->vpCBData);
    }
  }
  else
  {
    // no call backs
    ulExecuteOp(vpCtx, spChild);
  }
  
  // return the number of opcodes consumed by this operator
  return;
}

static void vSESTRNM(void* vpCtx, OP_INT_ALL* spOp)
{
  SESTWALKERCTX*  spCtx = (SESTWALKERCTX*)vpCtx;
  ulong           ulReturn = PFN_OK;
  ulong           ulContinue;
  INT_RULE_LIST*  spRule;
  OP_INT_ALL*     spFirst;
  PFN_SESTCB      pfnCB = spCtx->pfnaCallBacks[OP_INT_ID_RNM];
  
  // verify the operator type
  DASSERT(spOp->ulOpId == OP_INT_ID_RNM);
  
  // get the rule info
  spRule = &spCtx->spIntRules[spOp->ulRuleIndex];
  
  if(spRule->ulLeaf)
  {
    // the rule is already open, just do the call backs and treat as leaf node
    ++spCtx->ulTerminals;
    if(pfnCB)
    {
      ulReturn = pfnCB(SEM_PRE, spCtx->ulTreeDepth, LEAF_RULE, 0, spOp, spCtx->vpCBData);
      if(ulReturn == PFN_OK)
      {
        pfnCB(SEM_POST, spCtx->ulTreeDepth, LEAF_RULE, 0, spOp, spCtx->vpCBData);
      }
    }
  }
  else
  {
    // open the rule
    spRule->ulLeaf = APG_TRUE;

    // execute the first opcode
    if(pfnCB)
    {
      ulReturn = pfnCB(SEM_PRE, spCtx->ulTreeDepth, ROOT_RULE, 0, spOp, spCtx->vpCBData);
    }
    if(ulReturn == PFN_OK)
    {
      spFirst = &spCtx->spIntOpcodes[spRule->ulOpcodeOffset];
      
      // walk the rule
      ulContinue = ulExecuteOp(spCtx, spFirst);
      if(!ulContinue){goto MAXTERM_EXIT;}
      if(pfnCB)
      {
        pfnCB(SEM_POST, spCtx->ulTreeDepth, ROOT_RULE, 0, spOp, spCtx->vpCBData);
      }
    }

    // close the rule
    spRule->ulLeaf = APG_FALSE;
  }

  // exit on reaching maximum terminal leaf node count
MAXTERM_EXIT:
  return;
}

static void vSESTTRG(void* vpCtx, OP_INT_ALL* spOp)
{
  SESTWALKERCTX*  spCtx = (SESTWALKERCTX*)vpCtx;
  ulong           ulReturn = PFN_OK;
  PFN_SESTCB      pfnCB = spCtx->pfnaCallBacks[OP_INT_ID_TRG];

  // verify the operator type
  DASSERT(spOp->ulOpId == OP_INT_ID_TRG);
  
  ++spCtx->ulTerminals;
  if(pfnCB)
  {
    ulReturn = pfnCB(SEM_PRE, spCtx->ulTreeDepth, 0, 0, spOp, spCtx->vpCBData);
    if(ulReturn == PFN_OK)
    {
      pfnCB(SEM_POST, spCtx->ulTreeDepth, 0, 0, spOp, spCtx->vpCBData);
    }
  }
  
  // return the number of opcodes consumed by this operator
  return;
}

static void vSESTTBS(void* vpCtx, OP_INT_ALL* spOp)
{
  SESTWALKERCTX*  spCtx = (SESTWALKERCTX*)vpCtx;
  ulong           ulReturn = PFN_OK;
  PFN_SESTCB      pfnCB = spCtx->pfnaCallBacks[OP_INT_ID_TBS];
  
  // verify the operator type
  DASSERT(spOp->ulOpId == OP_INT_ID_TBS);

  ++spCtx->ulTerminals;
  if(pfnCB)
  {
    ulReturn = pfnCB(SEM_PRE, spCtx->ulTreeDepth, 0, 0, spOp, spCtx->vpCBData);
    if(ulReturn == PFN_OK)
    {
      pfnCB(SEM_POST, spCtx->ulTreeDepth, 0, 0, spOp, spCtx->vpCBData);
    }
  }
  
  // return the number of opcodes consumed by this operator
  return;
}

static void vSESTTLS(void* vpCtx, OP_INT_ALL* spOp)
{
  SESTWALKERCTX*  spCtx = (SESTWALKERCTX*)vpCtx;
  ulong           ulReturn = PFN_OK;
  PFN_SESTCB      pfnCB = spCtx->pfnaCallBacks[OP_INT_ID_TLS];
  
  // verify the operator type
  DASSERT(spOp->ulOpId == OP_INT_ID_TLS);
  
  ++spCtx->ulTerminals;
  if(pfnCB)
  {
    ulReturn = pfnCB(SEM_PRE, spCtx->ulTreeDepth, 0, 0, spOp, spCtx->vpCBData);
    if(ulReturn == PFN_OK)
    {
      pfnCB(SEM_POST, spCtx->ulTreeDepth, 0, 0, spOp, spCtx->vpCBData);
    }
  }
  
  // return the number of opcodes consumed by this operator
  return;
}

//////////////////////////////////////////////////////////////////////////
// PPT traversal node functions
//////////////////////////////////////////////////////////////////////////
static void vPPTREP(void* vpCtx, OP_INT_ALL* spOp)
{
  SESTWALKERCTX*  spCtx = (SESTWALKERCTX*)vpCtx;
  OP_INT_ALL*     spOperand;
  ulong           ulReturn = PFN_OK;
  ulong           ulOps;
  PFN_SESTCB      pfnCB = spCtx->pfnaCallBacks[OP_INT_ID_REP];
  DASSERT(spOp->ulOpId == OP_INT_ID_REP);
  DASSERT(!(spOp->ulMin == 0 && spOp->ulMax == 0));

  spOperand = spOp + 1;
  while(APG_TRUE)
  {
    if(pfnCB)
    {
      ulReturn = pfnCB(SEM_PRE, spCtx->ulTreeDepth, 0, 0, spOp, spCtx->vpCBData);
    }
    if(ulReturn == PFN_SKIP){break;}
    
    // execute the operand
    ulOps = ulExecuteOp(spCtx, spOperand);
    if(ulOps == 0){goto MAXTERM_EXIT;}
    
    if(pfnCB)
    {
      pfnCB(SEM_POST, spCtx->ulTreeDepth, 0, 0, spOp, spCtx->vpCBData);
    }
    break;
  }
  
  // exit on reaching maximum terminal leaf node count
MAXTERM_EXIT:
  return;
}

//////////////////////////////////////////////////////////////////////////
// DEFAULT CALL BACK FUNCTIONS
//////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////
// DESCRIPTION: Default RNM call back function. Just print the RNM and direction.
//
// FUNCTION: static ulong ulDefaultRNM
//
//      ARG: 1) ulong ulState
//              : the traversal state SEM_PRE, SEM_POST
//
//      ARG: 2) ulong ulTreeDepth
//              : the SEST tree depth
//
//      ARG: 3) ulong ulChildCount
//              : for ALT and CAT this is the child count (1-based)
//                0 indicates before and after all children have been traversed
//
//      ARG: 4) ulong ulCountMax
//              : number of children in the child list
//
//      ARG: 5) OP_INT_ALL* spOp
//              : the opcode that this SEST tree node represents
//
//      ARG: 6) void* vpData
//              : special data context
//
//   RETURN: PFN_OK is normal return, PFN_SKIP requests caller to skip the
//           traversal of the SEST branch below this node.
//
//////////////////////////////////////////////////////////////////////////
static ulong ulDefaultRNM(ulong ulState, ulong ulTreeDepth, ulong ulChildCount, ulong ulCountMax, OP_INT_ALL* spOp, void* vpData)
{
  switch(ulState)
  {
  case SEM_PRE:
    INDENT(ulTreeDepth);
    printf("RNM(SEM_PRE)\n");
    break;
  case SEM_POST:
    INDENT(ulTreeDepth);
    printf("RNM(SEM_POST)\n");
    break;
  default:
    DASSERT(APG_FALSE);
    break;
  }

  return PFN_OK;
}

static ulong ulDefaultALT(ulong ulState, ulong ulTreeDepth, ulong ulChildCount, ulong ulCountMax, OP_INT_ALL* spOp, void* vpData)
{
  switch(ulState)
  {
  case SEM_PRE:
    INDENT(ulTreeDepth);
    printf("ALT(SEM_PRE)\n");
    break;
  case SEM_POST:
    INDENT(ulTreeDepth);
    printf("ALT(SEM_POST)\n");
    break;
  default:
    DASSERT(APG_FALSE);
    break;
  }
  
  return PFN_OK;
}

static ulong ulDefaultCAT(ulong ulState, ulong ulTreeDepth, ulong ulChildCount, ulong ulCountMax, OP_INT_ALL* spOp, void* vpData)
{
  switch(ulState)
  {
  case SEM_PRE:
    INDENT(ulTreeDepth);
    printf("CAT(SEM_PRE)\n");
    break;
  case SEM_POST:
    INDENT(ulTreeDepth);
    printf("CAT(SEM_POST)\n");
    break;
  default:
    DASSERT(APG_FALSE);
    break;
  }
  
  return PFN_OK;
}

static ulong ulDefaultREP(ulong ulState, ulong ulTreeDepth, ulong ulChildCount, ulong ulCountMax, OP_INT_ALL* spOp, void* vpData)
{
  switch(ulState)
  {
  case SEM_PRE:
    INDENT(ulTreeDepth);
    printf("REP(SEM_PRE)\n");
    break;
  case SEM_POST:
    INDENT(ulTreeDepth);
    printf("REP(SEM_POST)\n");
    break;
  default:
    DASSERT(APG_FALSE);
    break;
  }
  
  return PFN_OK;
}

static ulong ulDefaultPRD(ulong ulState, ulong ulTreeDepth, ulong ulChildCount, ulong ulCountMax, OP_INT_ALL* spOp, void* vpData)
{
  switch(ulState)
  {
  case SEM_PRE:
    INDENT(ulTreeDepth);
    printf("PRD(SEM_PRE)\n");
    break;
  case SEM_POST:
    INDENT(ulTreeDepth);
    printf("PRD(SEM_POST)\n");
    break;
  default:
    DASSERT(APG_FALSE);
    break;
  }
  
  return PFN_OK;
}

static ulong ulDefaultTRG(ulong ulState, ulong ulTreeDepth, ulong ulChildCount, ulong ulCountMax, OP_INT_ALL* spOp, void* vpData)
{
  switch(ulState)
  {
  case SEM_PRE:
    INDENT(ulTreeDepth);
    printf("TRG(SEM_PRE)\n");
    break;
  case SEM_POST:
    INDENT(ulTreeDepth);
    printf("TRG(SEM_POST)\n");
    break;
  default:
    DASSERT(APG_FALSE);
    break;
  }
  
  return PFN_OK;
}

static ulong ulDefaultTBS(ulong ulState, ulong ulTreeDepth, ulong ulChildCount, ulong ulCountMax, OP_INT_ALL* spOp, void* vpData)
{
  switch(ulState)
  {
  case SEM_PRE:
    INDENT(ulTreeDepth);
    printf("TBS(SEM_PRE)\n");
    break;
  case SEM_POST:
    INDENT(ulTreeDepth);
    printf("TBS(SEM_POST)\n");
    break;
  default:
    DASSERT(APG_FALSE);
    break;
  }
  
  return PFN_OK;
}

static ulong ulDefaultTLS(ulong ulState, ulong ulTreeDepth, ulong ulChildCount, ulong ulCountMax, OP_INT_ALL* spOp, void* vpData)
{
  switch(ulState)
  {
  case SEM_PRE:
    INDENT(ulTreeDepth);
    printf("TLS(SEM_PRE)\n");
    break;
  case SEM_POST:
    INDENT(ulTreeDepth);
    printf("TLS(SEM_POST)\n");
    break;
  default:
    DASSERT(APG_FALSE);
    break;
  }
  
  return PFN_OK;
}

//////////////////////////////////////////////////////////////////////////
// DESCRIPTION: Prints a human-readable trace of the nodes visited.
//              WARNING: may generate a large amount of output.
//
// FUNCTION: void vTrace
//
//      ARG: 1) SESTWALKERCTX* spCtx
//              : pointer to a valid SESTWalker component context
//
//      ARG: 2) OP_INT_ALL* spOp
//              : pointer to the opcode being visited
//
//      ARG: 3) ulong ulUp
//              : direction of the traversal
//
//   RETURN: void
//
//////////////////////////////////////////////////////////////////////////
void vTrace(SESTWALKERCTX* spCtx, OP_INT_ALL* spOp, ulong ulUp)
{
  char* cpDirection = ulUp ? "up  " : "down";
  vIndent(spCtx->ulTreeDepth);
  switch(spOp->ulOpId)
  {
  case OP_INT_ID_ALT:
    printf("ALT(%s)[%lu]\n", cpDirection, spOp->ulCount);
    break;
  case OP_INT_ID_PRD:
    printf("PRD(%s)\n", cpDirection);
    break;
  case OP_INT_ID_CAT:
    printf("CAT(%s)[%lu]\n", cpDirection, spOp->ulCount);
    break;
  case OP_INT_ID_REP:
    {
      if(spOp->ulMax == UNDEFINED)
      {
        printf("REP(%s)[%lu, %s]\n", cpDirection, spOp->ulMin, "inf");
      }
      else
      {
        printf("REP(%s)[%lu, %lu]\n", cpDirection, spOp->ulMin, spOp->ulMax);
      }
    }
    break;
  case OP_INT_ID_RNM:
    {
      // get the rule name
      INT_RULE_LIST* spRule = &spCtx->spIntRules[spOp->ulRuleIndex];
      char* cpName = (char*)&spCtx->ucpStrTbl[spRule->ulNameOffset];
      if(spRule->ulLeaf)
      {
        if(ulUp)
        {
          printf("RNM[%s, %s, LEAF, terminal #%lu]\n",
            cpDirection, cpName, spCtx->ulTerminals);
        }
        else
        {
          printf("RNM[%s, %s, LEAF]\n",
            cpDirection, cpName);
        }
      }
      else
      {
        printf("RNM[%s, %s, ROOT]\n", cpDirection, cpName);
      }
    }
    break;
  case OP_INT_ID_TRG:
    {
      if(ulUp)
      {
        printf("TRG(%s)[%lu, %lu, terminal #%lu\n",
          cpDirection, spOp->ulMin, spOp->ulMax, spCtx->ulTerminals);
      }
      else
      {
        printf("TRG(%s)[%lu, %lu]\n", cpDirection, spOp->ulMin, spOp->ulMax);
      }
    }
    break;
  case OP_INT_ID_TBS:
    {
      if(ulUp)
      {
        printf("TBS(%s)[%lu, terminal #%lu]\n",
          cpDirection, spOp->ulLength, spCtx->ulTerminals);
      }
      else
      {
        printf("TBS(%s)[%lu]\n", cpDirection, spOp->ulLength);
      }
    }
    break;
  case OP_INT_ID_TLS:
    {
      if(ulUp)
      {
        printf("TLS(%s)[%lu, terminal #%lu]\n",
          cpDirection, spOp->ulLength, spCtx->ulTerminals);
      }
      else
      {
        printf("TLS(%s)[%lu]\n", cpDirection, spOp->ulLength);
      }
    }
    break;
  default:
    printf("ERR(%s)\n", cpDirection);
    break;
  }
}

