/********************************************************************
  APG - an ABNF Parser Generator
  Copyright  (C) 2007 Coast to Coast Research, Inc.
 
  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 2 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program.  If not, see
  <http://www.gnu.org/licenses/old-licenses/gpl-2.0.html>
  or write to the Free Software Foundation, Inc.,
  51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

    author: Lowell Thomas
            lowell@coasttocoastresearch.com
            http://www.coasttocoastresearch.com

   created: 8/26/2007 17:49
  filename: D:\Projects\Apgv5_0\Samples\IPv6\Timing.c
   purpose: Compare times for:
                - without PPT
                - with PPT
                - look ahead without PPT
                - look ahead with PPT
  
*********************************************************************/
#include "main.h"
#include <time.h>

static double dRunTiming(TESTDATA* spData);
static double dRunSimpleTiming(TESTDATA* spTestData);

void    vTestTiming(void* vpMemCtx, char* cpFileName)
{
  APG_BSTR          sSrc;
  void*             vpBstr;
  ulong             ulCount;
  TESTDATA          sData;
  double            dTimeSyntaxOnly;
  double            dTimeSyntaxOnlyPPT;
  double            dTimeLookAhead;
  double            dTimeLookAheadPPT;
  double            dTimeSimple;
  double            dFactor;
  
  // some valid and malformed IPv6 address
  sSrc.ulLen = ulGetFileSize(cpFileName);
  MAINASSERT(sSrc.ulLen);
  sSrc.ucpStr = (uchar*)vpMemAlloc(vpMemCtx, sSrc.ulLen + 10);
  MAINASSERT(sSrc.ucpStr);
  sSrc.ulLen = ulGetFile(cpFileName, (void*)sSrc.ucpStr);
  MAINASSERT(sSrc.ulLen);
  sSrc.ucpStr[sSrc.ulLen] = 0;
  
  // get the list of announcements
  vpBstr = vpVecCtor(vpMemCtx, sizeof(APG_BSTR), 0);
  MAINASSERT(vpBstr);
  ulCount = ulReadAddresses(&sSrc, vpBstr);
  MAINASSERT(ulCount);
  
  sData.vpMemCtx = vpMemCtx;
  sData.vpMsgs = vpBstr;
  sData.ulRepetitions = 500;

  printf("\n");
  printf("//////////////////////////////////////////////////////////////////////////\n");
  printf("// TIMING TESTS\n");
  printf("//////////////////////////////////////////////////////////////////////////\n");
  
  // syntax test without PPT
  sData.vpInit = vpIPv6SyntaxOnlyInit;
  sData.cpTitle = "Syntax Only";
  dTimeSyntaxOnly = dRunTiming(&sData);
  
  // look ahead test without PPT
  sData.vpInit = vpIPv6LookAheadInit;
  sData.cpTitle = "Look Ahead";
  dTimeLookAhead = dRunTiming(&sData);
  dFactor = dTimeSyntaxOnly / dTimeLookAhead;
  printf(" -          factor = %1.4g (improvement over Syntax Only)\n", dFactor);
  
  // syntax test with PPT
  sData.vpInit = vpIPv6SyntaxOnlyPPTInit;
  sData.cpTitle = "Syntax Only/PPT";
  dTimeSyntaxOnlyPPT = dRunTiming(&sData);
  dFactor = dTimeSyntaxOnly / dTimeSyntaxOnlyPPT;
  printf(" -          factor = %1.4g (improvement over Syntax Only)\n", dFactor);
  
  // look ahead test with PPT
  sData.vpInit = vpIPv6LookAheadPPTInit;
  sData.cpTitle = "look Ahead/PPT";
  dTimeLookAheadPPT = dRunTiming(&sData);
  dFactor = dTimeSyntaxOnly / dTimeLookAheadPPT;
  printf(" -          factor = %1.4g (improvement over Syntax Only)\n", dFactor);

  // simple hand-coded parsing
  sData.vpInit = NULL;
  sData.cpTitle = "Simple Hand Coding";
  dTimeSimple = dRunSimpleTiming(&sData);
  dFactor = dTimeSyntaxOnly / dTimeSimple;
  printf(" -          factor = %1.4g (improvement over Syntax Only)\n", dFactor);
  
  // clean up resources
  vVecDtor(vpBstr);
  vMemFree(vpMemCtx, sSrc.ucpStr);
}

static double dRunTiming(TESTDATA* spTestData)
{
  double            dReturn = 0.0;
  double            dStart;
  double            dFinish;
  double            dClocksPerSec;
  double            dDuration;
  APG_BSTR*         spBstr;
  ulong             ulTest;
  ulong             ulCount;
  ulong             ulAdd;
  ulong             ulRep;
  ulong             ulAddresses;
  ulong             ulCharacters;
  void*             vpParserCtx;
  DASSERT(spTestData);
  DASSERT(spTestData->vpMemCtx);
  DASSERT(spTestData->vpMsgs);
  DASSERT(spTestData->vpInit);
  
  memset((void*)&spTestData->sStats, 0, sizeof(spTestData->sStats));
  if(spTestData->cpTitle && *spTestData->cpTitle != 0)
  {
    // print a header
    printf("\n");
    printf("%s\n", spTestData->cpTitle);
  }
  
  // get the list of addresses
  spBstr = (APG_BSTR*)vpVecFront(spTestData->vpMsgs);
  MAINASSERT(spBstr);
  ulCount = ulVecSize(spTestData->vpMsgs);
  MAINASSERT(ulCount);
  
  // initialize the parser context
  ulTest = ulParserInit(spTestData->vpInit, spTestData->vpMemCtx);
  MAINASSERT(ulTest);
  vpParserCtx = vpParserCtor(spTestData->vpInit, spTestData->vpMemCtx);
  MAINASSERT(vpParserCtx);
  
  // translate each address
  ulAddresses = 0;
  ulCharacters = 0;
  dClocksPerSec = (double)CLOCKS_PER_SEC;
  dStart = (double)clock();
  for(ulRep = 0; ulRep < spTestData->ulRepetitions; ++ulRep)
  {
    spBstr = (APG_BSTR*)vpVecFront(spTestData->vpMsgs);
    MAINASSERT(spBstr);
    for(ulAdd = 0; ulAdd < ulCount; ++ulAdd, ++spBstr)
    {
      ++ulAddresses;
      ulCharacters += spBstr->ulLen;
      ulTest = ulParserSyntaxAnalysis(vpParserCtx, 0, spBstr->ucpStr, spBstr->ulLen, spBstr, APG_FALSE);
      MAINASSERT(ulTest);
//       ulTest = ulParserState(vpParserCtx, &sState);
//       MAINASSERT(ulTest && sState.ulState == MATCH);
//       ulTest = ulParserStats(vpParserCtx, &sStats);
//       MAINASSERT(ulTest);
//       
//       // collective statistics
//       ulpTotalStats = (ulong*)&spTestData->sStats;
//       ulpStats = (ulong*)&sStats;
//       for(ulTest = 0; ulTest < sizeof(sStats)/sizeof(ulong); ++ulTest)
//       {
//         ulpTotalStats[ulTest] += ulpStats[ulTest];
//       }
    }
  }
  
  dFinish = (double)clock();
  dDuration = ((dFinish - dStart) * 1000.0) / dClocksPerSec;
  spTestData->dAddressPerMsec = (double)ulAddresses / dDuration;
  spTestData->dCharactersPerMsec = (double)ulCharacters / dDuration;

  printf(" -       Addresses = %lu\n", ulAddresses);
  printf(" -      Characters = %lu\n", ulCharacters);
  printf(" -      Time(msec) = %1.4g\n", dDuration);
  printf(" -  Addresses/msec = %1.4g\n", spTestData->dAddressPerMsec);
  printf(" - Characters/msec = %1.4g\n", spTestData->dCharactersPerMsec);
  

  // clean up
  vParserDtor(vpParserCtx);
  
  dReturn = dDuration;
  return dReturn;
}

static double dRunSimpleTiming(TESTDATA* spTestData)
{
  double            dReturn = 0.0;
  double            dStart;
  double            dFinish;
  double            dClocksPerSec;
  double            dDuration;
  APG_BSTR*         spBstr;
  ulong             ulCount;
  ulong             ulAdd;
  ulong             ulRep;
  ulong             ulAddresses;
  ulong             ulCharacters;
  ushort            uspAddress[IPV6_SHORTS];
  DASSERT(spTestData);
  DASSERT(spTestData->vpMsgs);
  
  memset((void*)&spTestData->sStats, 0, sizeof(spTestData->sStats));
  if(spTestData->cpTitle && *spTestData->cpTitle != 0)
  {
    // print a header
    printf("\n");
    printf("%s\n", spTestData->cpTitle);
  }
  
  // get the list of addresses
  spBstr = (APG_BSTR*)vpVecFront(spTestData->vpMsgs);
  MAINASSERT(spBstr);
  ulCount = ulVecSize(spTestData->vpMsgs);
  MAINASSERT(ulCount);
  
  // translate each address
  ulAddresses = 0;
  ulCharacters = 0;
  dClocksPerSec = (double)CLOCKS_PER_SEC;
  dStart = (double)clock();
  for(ulRep = 0; ulRep < spTestData->ulRepetitions; ++ulRep)
  {
    spBstr = (APG_BSTR*)vpVecFront(spTestData->vpMsgs);
    MAINASSERT(spBstr);
    for(ulAdd = 0; ulAdd < ulCount; ++ulAdd, ++spBstr)
    {
      ++ulAddresses;
      ulCharacters += spBstr->ulLen;
      ulIPv6StringToBinary(spBstr, uspAddress, NULL);
    }
  }
  
  dFinish = (double)clock();
  dDuration = ((dFinish - dStart) * 1000.0) / dClocksPerSec;
  spTestData->dAddressPerMsec = (double)ulAddresses / dDuration;
  spTestData->dCharactersPerMsec = (double)ulCharacters / dDuration;
  
  printf(" -       Addresses = %lu\n", ulAddresses);
  printf(" -      Characters = %lu\n", ulCharacters);
  printf(" -      Time(msec) = %1.4g\n", dDuration);
  printf(" -  Addresses/msec = %1.4g\n", spTestData->dAddressPerMsec);
  printf(" - Characters/msec = %1.4g\n", spTestData->dCharactersPerMsec);

  dReturn = dDuration;
  return dReturn;
}
