/********************************************************************
  APG - an ABNF Parser Generator
  Copyright  (C) 2007 Coast to Coast Research, Inc.
 
  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 2 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program.  If not, see
  <http://www.gnu.org/licenses/old-licenses/gpl-2.0.html>
  or write to the Free Software Foundation, Inc.,
  51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

    author: Lowell Thomas
            lowell@coasttocoastresearch.com
            http://www.coasttocoastresearch.com

   created: 8/27/2007 10:14
  filename: D:\Projects\Apgv5_0\Samples\SDP\Connection.c
   purpose: Parse the IPv4 & IPv6 addresses in the connection field.

                Illustrates:
                1) Use of Catalog to rapidly find a lower-level AST
                   node (phrase) without traversing the entire tree.
                2) Partial traversal of the AST, starting with
                   a node lower than the start rule root node.
                3) Writing multiple translations within the parser's
                   call back functions.
                4) Using an external, hand-written parser for the final
                   interpretation of the IP address string.

  NOTES: The Catalog is the quickest and most convenient way to find
         lower-level AST nodes without having to do a complete
         traversal of the AST. The Catalog also provides the means
         to travel up and down the AST from a given node which is
         convenient when only one or two levels up or down is needed.
         However, for data buried deep below a given node, a AST
         traversal is again more convenient. The Catalog therefore, also
         provides the ability to traverse a partial AST, beginning
         at an arbitrary node - not just the root node. A combination of
         horizontal and vertical traversal of the AST can be done.
  
*********************************************************************/
#include "main.h"

static ulong ulPrintAddress(void* vpCatalogCtx);

//////////////////////////////////////////////////////////////////////////
// DESCRIPTION: Driver for the tests
//
// FUNCTION: ulong ulConnectionTest
//
//      ARG: 1) APG_BSTR* spAnnouncement
//              : binary string defining the list of announcements
//
//      ARG: 2) void* vpMemCtx
//              : pointer to a valid memory component context
//
//      ARG: 3) void* vpParserCtx
//              : pointer to a valid parser component context
//
//   RETURN: true on success
//
//////////////////////////////////////////////////////////////////////////
ulong ulConnectionTest(APG_BSTR* spAnnouncement, void* vpMemCtx, void* vpParserCtx)
{
  ulong   ulReturn = APG_FALSE;
  APG_PARSER_STATE sState;
  LINE_ITERATOR sIter;
  ulong ulTest;
  APG_BSTR sLine;
  ulong ulID;
  void* vpCatalogCtx;
  ulong ulPhraseCount;
  
  // iterate over all announcement lines
  ulTest = ulLinesInit(&sIter, spAnnouncement);
  SDPASSERT(ulTest);
  while(APG_TRUE)
  {
    ulTest = ulLinesNext(&sIter, &sLine, &ulID);
    SDPASSERT(ulTest);
    if(sLine.ucpStr == NULL)
    {
      // end of announcement
      break;
    }
    
    // find the session-section connection field
    if(ulID == SESSION_CONNECTION)
    {
      // parse the session-section connection line only
      ulTest = ulParserSyntaxAnalysis(vpParserCtx, RULE_SDP_CONNECTION_FIELD, sLine.ucpStr, sLine.ulLen, NULL, APG_FALSE);
      SDPASSERT(ulTest);
      ulTest = ulParserState(vpParserCtx, &sState);
      SDPASSERT(ulTest);
      if(sState.ulState == MATCH)
      {
        // catalog the phrases in the connection field
        vpCatalogCtx = vpCatalogCtor(vpParserCtx);
        SDPASSERT(vpCatalogCtx);
        ulPhraseCount = ulCatalog(vpCatalogCtx);
        SDPASSERT(ulPhraseCount);

        // find and print the IP address
        ulTest = ulPrintAddress(vpCatalogCtx);
        SDPASSERT(ulTest);
      }
      
      break;
    }
  }
  
  // success
  ulReturn = APG_TRUE;
  return ulReturn;
}

//////////////////////////////////////////////////////////////////////////
// DESCRIPTION: Find the address type and print it
//              Addresses may be IPv6 uni/multicast or IPv4 uni/multicast
//
// FUNCTION: static ulong ulPrintAddress
//
//      ARG: 1) void* vpCatalogCtx
//              : pointer to a valid catalog component context
//
//   RETURN: true on success
//
//////////////////////////////////////////////////////////////////////////
static ulong ulPrintAddress(void* vpCatalogCtx)
{
  ulong   ulReturn = APG_FALSE;
  APG_CAT_ITERATOR sIter;
  APG_CAT_ITERATOR* spIter;
  ulong ulTest;
  ulong ulIsMulticast;
  char    caIPv4String[IPV4_MAXLEN + 1];
  char    caIPv6String[IPV6_MAXLEN + 1];
  ushort  usaIPv6[IPV6_SHORTS];
  char    caIPv4[IPV4_OCTETS];
  
  // single-point-of-exit loop
  while(APG_TRUE)
  {
    // only one IP address of one type will be present
    // try them all until we find the one that is present
    spIter = spCatalogGetPhraseList(vpCatalogCtx, RULE_SDP_IPV4UNICAST, &sIter);
    if(spIter)
    {
      // parse the string as IPv4 unicast
      ulTest = ulIPv4StringToBinary(&spIter->sPhrase, caIPv4, &ulIsMulticast);
      if(!ulTest){goto APG_EXIT;}
      if(ulIsMulticast){goto APG_EXIT;}
      ulTest = ulIPv4BinaryToString(caIPv4, caIPv4String);
      if(!ulTest){goto APG_EXIT;}
      printf("\n");
      printf("*** IPv4 Unicast - Normal Form Address\n");
      printf("%s\n", caIPv4String);
      
      // success
      ulReturn = APG_TRUE;
      break;
    }
    
    spIter = spCatalogGetPhraseList(vpCatalogCtx, RULE_SDP_IPV4MULTICAST, &sIter);
    if(spIter)
    {
      // parse the string as IPv4 multicast
      ulTest = ulIPv4StringToBinary(&spIter->sPhrase, caIPv4, &ulIsMulticast);
      if(!ulTest){goto APG_EXIT;}
      if(!ulIsMulticast){goto APG_EXIT;}
      ulTest = ulIPv4BinaryToString(caIPv4, caIPv4String);
      if(!ulTest){goto APG_EXIT;}
      printf("\n");
      printf("*** IPv4 Multicast - Normal Form Address\n");
      printf("%s\n", caIPv4String);
      
      // success
      ulReturn = APG_TRUE;
      break;
    }
    
    spIter = spCatalogGetPhraseList(vpCatalogCtx, RULE_SDP_IPV6UNICAST, &sIter);
    if(spIter)
    {
      // parse the string as IPv6 unicast
      ulTest = ulIPv6StringToBinary(&spIter->sPhrase, usaIPv6, &ulIsMulticast);
      if(!ulTest){goto APG_EXIT;}
      if(ulIsMulticast){goto APG_EXIT;}
      ulTest = ulIPv6BinaryToString(usaIPv6, caIPv6String);
      if(!ulTest){goto APG_EXIT;}
      printf("\n");
      printf("*** IPv6 Unicast - Normal Form Address\n");
      printf("%s\n", caIPv6String);
      
      // success
      ulReturn = APG_TRUE;
      break;
    }
    
    spIter = spCatalogGetPhraseList(vpCatalogCtx, RULE_SDP_IPV6MULTICAST, &sIter);
    if(spIter)
    {
      // parse the string as IPv6 multicast
      ulTest = ulIPv6StringToBinary(&spIter->sPhrase, usaIPv6, &ulIsMulticast);
      if(!ulTest){goto APG_EXIT;}
      if(!ulIsMulticast){goto APG_EXIT;}
      ulTest = ulIPv6BinaryToString(usaIPv6, caIPv6String);
      if(!ulTest){goto APG_EXIT;}
      printf("\n");
      printf("*** IPv6 Multicast - Normal Form Address\n");
      printf("%s\n", caIPv6String);
      
      // success
      ulReturn = APG_TRUE;
      break;
    }
    
    // IP address not found
    printf("\n");
    printf("*** <error - no IP Address found>\n");
    break;
  } // end of single-point-of-exit loop
  
APG_EXIT:
  return ulReturn;
}
