/********************************************************************
  APG - an ABNF Parser Generator
  Copyright  (C) 2007 Coast to Coast Research, Inc.
 
  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 2 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program.  If not, see
  <http://www.gnu.org/licenses/old-licenses/gpl-2.0.html>
  or write to the Free Software Foundation, Inc.,
  51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

    author: Lowell Thomas
            lowell@coasttocoastresearch.com
            http://www.coasttocoastresearch.com

   created: 8/28/2007 9:37
  filename: D:\Projects\Apgv5_0\Samples\SIP\Timing.c
   purpose: Timing test functions.
  
*********************************************************************/
#include "main.h"
#include <time.h>

typedef ulong (*PFN_PARSER)(APG_BSTR* spMsg, void* vpParserCtx);
typedef struct
{
  void*         vpLhsCtx;
  void*         vpRhsCtx;
  PFN_PARSER    pfnLhsParser;
  PFN_PARSER    pfnRhsParser;
  APG_BSTR*     spMsg;
  ulong         ulMsgCount;
  ulong         ulReps;
} TIMING_INPUT;

typedef struct
{
  // msg totals
  ulong   ulMsgs;
  ulong   ulChars;

  // lhs results
  ulong   ulLhsTime;
  double  dLhsTimeMsec;
  double  dLhsMsgsPerMsec;
  double  dLhsCharsPerMsec;

  // rhs results
  ulong   ulRhsTime;
  double  dRhsTimeMsec;
  double  dRhsMsgsPerMsec;
  double  dRhsCharsPerMsec;
} TIMING_OUTPUT;

static ulong ulFullParseWithCatalog(APG_BSTR* spMsg, void* vpParserCtx);
static ulong ulFullParseWithoutCatalog(APG_BSTR* spMsg, void* vpParserCtx);
static ulong ulLazyParse(APG_BSTR* spMsg, void* vpParserCtx);
static ulong ulCompareTimes(TIMING_INPUT* spIn, TIMING_OUTPUT* spOut);
static void  vPrintResults(TIMING_OUTPUT* spOut, char* cpLhsLabel, char* cpRhsLabel);

//////////////////////////////////////////////////////////////////////////
// DESCRIPTION: Timing test, comparing times for parsing SIP with and without PPT and
//              SIP lazy parsing with and without PPT,
//
// FUNCTION: ulong ulTimingTest
//
//      ARG: 1) void* vpMsgs
//              : vector of binary strings defining a list of SIP messages
//
//      ARG: 2) ulong ulRepetitions
//              : number of times to repeat parsing all messages in the list
//
//   RETURN: true on success
//
//////////////////////////////////////////////////////////////////////////
ulong ulTimingTest(void* vpMsgs, ulong ulRepetitions)
{
  ulong         ulReturn = APG_FALSE;
  APG_BSTR*     spMsg;
  TIMING_INPUT  sInput;
  TIMING_OUTPUT sOutput;
  ulong         ulCount;
  ulong         ulTest;
  DASSERT(vpMsgs);

  spMsg = (APG_BSTR*)vpVecFront(vpMsgs);
  SIPASSERT(spMsg);
  ulCount = ulVecSize(vpMsgs);
  SIPASSERT(ulCount);

  // compare SIP without PPT and with PPT
  sInput.vpLhsCtx = vpParserCtor(vpSIPInit, vpMemCtx);
  SIPASSERT(sInput.vpLhsCtx);
  sInput.vpRhsCtx = vpParserCtor(vpSIPPPTInit, vpMemCtx);
  SIPASSERT(sInput.vpRhsCtx);
  sInput.pfnLhsParser = ulFullParseWithoutCatalog;
  sInput.pfnRhsParser = ulFullParseWithoutCatalog;
  sInput.spMsg = spMsg;
  sInput.ulMsgCount = ulCount;
  sInput.ulReps = ulRepetitions;
  ulTest = ulCompareTimes(&sInput, &sOutput);
  SIPASSERT(ulTest);

  // display results
  vPrintResults(&sOutput, "w/out PPT", "w/PPT");
  vParserDtor(sInput.vpLhsCtx);
  vParserDtor(sInput.vpRhsCtx);

  // compare full grammar parse and lazy parse (both with PPT)
  sInput.vpLhsCtx = vpParserCtor(vpSIPPPTInit, vpMemCtx);
  SIPASSERT(sInput.vpLhsCtx);
  sInput.vpRhsCtx = vpParserCtor(vpSIPPPTLazyInit, vpMemCtx);
  SIPASSERT(sInput.vpRhsCtx);
  sInput.pfnLhsParser = ulFullParseWithoutCatalog;
  sInput.pfnRhsParser = ulLazyParse;
  sInput.spMsg = spMsg;
  sInput.ulMsgCount = ulCount;
  sInput.ulReps = ulRepetitions;
  ulTest = ulCompareTimes(&sInput, &sOutput);
  SIPASSERT(ulTest);
  
  // display results
  vPrintResults(&sOutput, "full parse", "lazy parse");
  vParserDtor(sInput.vpLhsCtx);
  vParserDtor(sInput.vpRhsCtx);
  
  // compare SIP with cataloging and without cataloging (both with PPT)
  sInput.vpLhsCtx = vpParserCtor(vpSIPPPTInit, vpMemCtx);
  SIPASSERT(sInput.vpLhsCtx);
  sInput.vpRhsCtx = vpParserCtor(vpSIPPPTInit, vpMemCtx);
  SIPASSERT(sInput.vpRhsCtx);
  sInput.pfnLhsParser = ulFullParseWithCatalog;
  sInput.pfnRhsParser = ulFullParseWithoutCatalog;
  sInput.spMsg = spMsg;
  sInput.ulMsgCount = ulCount;
  sInput.ulReps = ulRepetitions;
  ulTest = ulCompareTimes(&sInput, &sOutput);
  SIPASSERT(ulTest);
  
  // display results
  vPrintResults(&sOutput, "w/catalog", "w/out catalog");
  vParserDtor(sInput.vpLhsCtx);
  vParserDtor(sInput.vpRhsCtx);
  
  // success
  ulReturn = APG_TRUE;

APG_EXIT:
  return ulReturn;
}

//////////////////////////////////////////////////////////////////////////
// DESCRIPTION: Runs two different parses over the same set of messages
//              and compares the two timing results
//
// FUNCTION: static ulong ulCompareTimes
//
//      ARG: 1) TIMING_INPUT* spIn
//              : defines the two parsers to compare
//
//      ARG: 2) TIMING_OUTPUT* spOut
//              : on return, gives the timing comparison results
//
//   RETURN: true on success
//
//////////////////////////////////////////////////////////////////////////
static ulong ulCompareTimes(TIMING_INPUT* spIn, TIMING_OUTPUT* spOut)
{
  ulong     ulReturn = APG_FALSE;
  APG_BSTR* spMsg;
  APG_BSTR  sHdrs;
  APG_BSTR  sBody;
  ulong     ulMsgIndex;
  ulong     ulRepIndex;
  ulong     ulMsgs;
  ulong     ulReps;
  ulong     ulCharCount;
  ulong     ulMsgCount;
  ulong     ulTest;
  double    dStartTime;
  double    dEndTime;
  double    dDuration;
  double    dClocksPerSec = (double)CLOCKS_PER_SEC;

  // debug validation
  DASSERT(spIn);
  DASSERT(spOut);

  // general set up
  ulMsgs = spIn->ulMsgCount;
  ulReps = spIn->ulReps;

  // lhs setup
  ulMsgCount = 0;
  ulCharCount = 0;
  dStartTime = (double)clock();

  // loop over repetitions
  for(ulRepIndex = 0; ulRepIndex < ulReps; ++ulRepIndex)
  {
    // loop over all msgs
    spMsg = spIn->spMsg;
    for(ulMsgIndex = 0; ulMsgIndex < ulMsgs; ++ulMsgIndex, ++spMsg)
    {
      ulTest = ulHdrQuickScan(spMsg, &sHdrs, &sBody);
      SIPASSERT(ulTest);
      ++ulMsgCount;
      ulCharCount += sHdrs.ulLen;

      // left-hand side parse of entire header section
      ulTest = spIn->pfnLhsParser(&sHdrs, spIn->vpLhsCtx);
      SIPASSERT(ulTest);
    }
  }
  
  dEndTime = (double)clock();
  dDuration = ((dEndTime - dStartTime) * 1000.0) / dClocksPerSec;
  spOut->ulMsgs = ulMsgCount;
  spOut->ulChars = ulCharCount;
  spOut->dLhsTimeMsec = dDuration;
  spOut->dLhsMsgsPerMsec = (double)ulMsgCount / dDuration;
  spOut->dLhsCharsPerMsec = (double)ulCharCount / dDuration;

  // rhs setup
  ulMsgCount = 0;
  ulCharCount = 0;
  dStartTime = (double)clock();
  
  // loop over repetitions
  for(ulRepIndex = 0; ulRepIndex < ulReps; ++ulRepIndex)
  {
    // loop over all msgs
    spMsg = spIn->spMsg;
    for(ulMsgIndex = 0; ulMsgIndex < ulMsgs; ++ulMsgIndex, ++spMsg)
    {
      ulTest = ulHdrQuickScan(spMsg, &sHdrs, &sBody);
      SIPASSERT(ulTest);
      ++ulMsgCount;
      ulCharCount += sHdrs.ulLen;
      
      // right-hand side parse of entire header section
      ulTest = spIn->pfnRhsParser(&sHdrs, spIn->vpRhsCtx);
      SIPASSERT(ulTest);
    }
  }
  
  dEndTime = (double)clock();
  dDuration = ((dEndTime - dStartTime) * 1000.0) / dClocksPerSec;
  spOut->dRhsTimeMsec = dDuration;
  spOut->dRhsMsgsPerMsec = (double)ulMsgCount / dDuration;
  spOut->dRhsCharsPerMsec = (double)ulCharCount / dDuration;
  // success
  ulReturn = APG_TRUE;
  
APG_EXIT:
  return ulReturn;
}

//////////////////////////////////////////////////////////////////////////
// DESCRIPTION: Does a full parse of a SIP message including catalog generation.
//
// FUNCTION: static ulong ulFullParseWithCatalog
//
//      ARG: 1) APG_BSTR* spMsg
//              : binary string defining the SIP message
//
//      ARG: 2) void* vpParserCtx
//              : pointer to a SIP parser data definition
//                (as found in a SIP parser header file)
//
//   RETURN: true on success
//
//////////////////////////////////////////////////////////////////////////
static ulong ulFullParseWithCatalog(APG_BSTR* spMsg, void* vpParserCtx)
{
  ulong             ulReturn = APG_FALSE;
  ulong             ulTrace = APG_FALSE;
  ulong             ulTest;
  void*             vpData = NULL;
  void*             vpCatalogCtx;
  APG_PARSER_STATE  sState;

  // syntax analysis
  ulTest = ulParserSyntaxAnalysis(vpParserCtx, 0, spMsg->ucpStr, spMsg->ulLen, vpData, ulTrace);
  SIPASSERT(ulTest);
  ulTest = ulParserState(vpParserCtx, &sState);
  SIPASSERT(ulTest);
  if(sState.ulState == NOMATCH)
  {
    SIPASSERT(APG_FALSE);
  }
  
  if(sState.ulState == MATCH)
  {
    // catalog the phrases
    vpCatalogCtx = vpCatalogCtor(vpParserCtx);
    SIPASSERT(vpCatalogCtx);
    ulTest = ulCatalog(vpCatalogCtx);
    SIPASSERT(ulTest);
  }

  // success
  ulReturn = APG_TRUE;

APG_EXIT:
  return ulReturn;
}

//////////////////////////////////////////////////////////////////////////
// DESCRIPTION: Does a full parse of a SIP message without generating a catalog.
//
// FUNCTION: static ulong ulFullParseWithoutCatalog
//
//      ARG: 1) APG_BSTR* spMsg
//              : binary string defining the SIP message
//
//      ARG: 2) void* vpParserCtx
//              : pointer to a SIP parser data definition
//                (as found in a SIP parser header file)
//
//   RETURN: true on success
//
//////////////////////////////////////////////////////////////////////////
static ulong ulFullParseWithoutCatalog(APG_BSTR* spMsg, void* vpParserCtx)
{
  ulong             ulReturn = APG_FALSE;
  ulong             ulTrace = APG_FALSE;
  ulong             ulTest;
  void*             vpData = NULL;
  APG_PARSER_STATE  sState;
  
  // syntax analysis
  ulTest = ulParserSyntaxAnalysis(vpParserCtx, 0, spMsg->ucpStr, spMsg->ulLen, vpData, ulTrace);
  SIPASSERT(ulTest);
  ulTest = ulParserState(vpParserCtx, &sState);
  SIPASSERT(ulTest);
  if(sState.ulState == NOMATCH)
  {
    SIPASSERT(APG_FALSE);
  }
  
  // success
  ulReturn = APG_TRUE;
  
APG_EXIT:
  return ulReturn;
}

//////////////////////////////////////////////////////////////////////////
// DESCRIPTION: Does a "lazy" parse of a SIP message.
//              Start line is hand-coded parsed.
//              Individual message headers are APG parsed.
//
// FUNCTION: static ulong ulLazyParse
//
//      ARG: 1) APG_BSTR* spMsg
//              : binary string defining the SIP message
//
//      ARG: 2) void* vpParserCtx
//              : pointer to a SIP parser data definition
//                (as found in a SIP parser header file)
//
//   RETURN: true on success
//
//////////////////////////////////////////////////////////////////////////
static ulong ulLazyParse(APG_BSTR* spMsg, void* vpParserCtx)
{
  ulong         ulReturn = APG_FALSE;
  APG_BSTR      sMsgCopy;
  void*         vpHdrCtx;
  ulong         ulTrace = APG_FALSE;
  void*         vpData = NULL;
  ulong         ulStartRule;
  ulong         ulTest;
  HDR_ITERATOR  sIter;
  ulong         ulID;
  ulong         ulType;
  SIP_RESPONSE  sResponse;
  SIP_REQUEST   sRequest;
  APG_PARSER_STATE  sState;

  // get a header catalog context
  vpHdrCtx = vpHdrCtor(vpMemCtx);
  SIPASSERT(vpHdrCtx);

  sMsgCopy = *spMsg;
  ulTest = ulHdrCatalog(vpHdrCtx, &sMsgCopy);
  SIPASSERT(ulTest);
  
  // parse the start line
  ulTest = ulHdrStartLine(vpHdrCtx, &ulType, &sRequest, &sResponse);
  SIPASSERT(ulTest);

  // parse the headers
  for(ulID = 0; ulID < HDR_COUNT; ++ulID)
  {
    ulTest = ulHdrIterInit(vpHdrCtx, &sIter, ulID);
    if(sIter.ulCount > 0)
    {
      ulong ul;
      for(ul = 0; ul < sIter.ulCount; ++ul)
      {
        // parse the header value parameters
        ulTest = ulHdrIterIndex(&sIter, ul);
        SIPASSERT(ulTest);
        if(sIter.spValue->ulLen == 0)
        {
          // skip empty header values
          continue;
        }
        
        // get the start rule
        ulStartRule = ulGetStartRule(ulID);
        SIPASSERT(ulStartRule != UNDEFINED);
        
        // syntax analysis
        ulTest = ulParserSyntaxAnalysis(vpParserCtx, ulStartRule, sIter.spValue->ucpStr, sIter.spValue->ulLen, vpData, ulTrace);
        SIPASSERT(ulTest);
        ulTest = ulParserState(vpParserCtx, &sState);
        SIPASSERT(ulTest);
        if(sState.ulState == NOMATCH)
        {
          SIPASSERT(APG_FALSE);
        }
      }
    }
  }

  // success
  ulReturn = APG_TRUE;

APG_EXIT:
  return ulReturn;
}

//////////////////////////////////////////////////////////////////////////
// DESCRIPTION: Formats and prints the results for comparision of timing
//              results for two different parsers
//
// FUNCTION: static void vPrintResults
//
//      ARG: 1) TIMING_OUTPUT* spOut
//              : timing information
//
//      ARG: 2) char* cpLhsLabel
//              : text label for the left-hand side of comparison
//
//      ARG: 3) char* cpRhsLabel
//              : text label for the right-hand side of comparison
//
//   RETURN: void
//
//////////////////////////////////////////////////////////////////////////
static void  vPrintResults(TIMING_OUTPUT* spOut, char* cpLhsLabel, char* cpRhsLabel)
{
  double dFactor;

  // print Lhs results
  printf("\n");
  printf("%s\n", cpLhsLabel);
  printf(" -        Messages = %lu\n", spOut->ulMsgs);
  printf(" -      Characters = %lu\n", spOut->ulChars);
  printf(" -      Time(msec) = %1.4g\n", spOut->dLhsTimeMsec);
  printf(" -   Messages/msec = %1.4g\n", spOut->dLhsMsgsPerMsec);
  printf(" - Characters/msec = %1.4g\n", spOut->dLhsCharsPerMsec);

  // print Rhs results
  printf("\n");
  printf("%s\n", cpRhsLabel);
  printf(" -        Messages = %lu\n", spOut->ulMsgs);
  printf(" -      Characters = %lu\n", spOut->ulChars);
  printf(" -      Time(msec) = %1.4g\n", spOut->dRhsTimeMsec);
  printf(" -   Messages/msec = %1.4g\n", spOut->dRhsMsgsPerMsec);
  printf(" - Characters/msec = %1.4g\n", spOut->dRhsCharsPerMsec);

  // print delta factor
  dFactor = spOut->dLhsTimeMsec / spOut->dRhsTimeMsec;
  printf("\n");
  printf("Time Factor (%s = factor x %s)\n", cpLhsLabel, cpRhsLabel);
  printf(" -          factor = %1.4g\n", dFactor);
  
}
