/*! \file Config.cpp
 <pre>
 <b>File:</b> Config.cpp
 <b>Project:</b> Robocup 3D
 <b>Authors:</b> Hviezdna Jedenastka
 <b>Created:</b> 03/10/2008
 <b>Contents:</b> <TEXT>
 </pre>
 */

#include <iostream>
#include "tinyxml.h"
#include "Config.h"
#include "Logger.h"

using namespace std;

PlayerConfig::PlayerConfig() {
    
    team[0] = '\0';
    number = 0;
}

WorldConfig::WorldConfig() {
    
    ballRadius = 0.0f;
}

Config::Config() {
    
    useThreads = false;
    strcpy(serverIP, "127.0.0.1");
    serverPort = 3100;
    connectionType = CONN_TCP;
}

bool Config::getAttributeString(TiXmlElement * pElement, const char *pAttribute, char *string, unsigned int stringSize) {
    
    if (!pElement)
        return 0;
    
    TiXmlAttribute *pAttrib = pElement->FirstAttribute();
    
    while (pAttrib) {
        
        if (!strcasecmp(pAttrib->Name(), pAttribute)) {
            
            strncpy(string, pAttrib->Value(), stringSize - 1);
            string[stringSize - 1] = '\0';
            return true;
        }
        
        pAttrib = pAttrib->Next();
    }
    
    return false;
}

bool Config::getAttributeInt(TiXmlElement * pElement, const char *pAttribute, int &integer) {
    
    if (!pElement)
        return 0;
    
    TiXmlAttribute *pAttrib = pElement->FirstAttribute();
    
    while (pAttrib) {
        
        if (!strcasecmp(pAttrib->Name(), pAttribute)) {
            
            if (pAttrib->QueryIntValue(&integer) == TIXML_SUCCESS)
                return true;
        }
        
        pAttrib = pAttrib->Next();
    }
    
    return false;
}

bool Config::getAttributeDouble(TiXmlElement * pElement, const char *pAttribute, double &num) {
    
    if (!pElement)
        return 0;
    
    TiXmlAttribute *pAttrib = pElement->FirstAttribute();
    
    while (pAttrib) {
        
        if (!strcasecmp(pAttrib->Name(), pAttribute)) {
            
            if (pAttrib->QueryDoubleValue(&num) == TIXML_SUCCESS)
                return true;
        }
        
        pAttrib = pAttrib->Next();
    }
    
    return false;
}

int Config::loadConfig(const char *xmlFile) {
    
    if (!xmlDocument.LoadFile(xmlFile)) {
        
#ifdef _LOGGER
        Logger::logoutput(Logger::LOG_CONFIG, log_config_failed, xmlFile);
#endif
        
        cerr << "Failed to open config file " << xmlFile << endl;
        return 1;
    }
    
    TiXmlNode *xmlNode = xmlDocument.FirstChild("Sirius");
    
    if (xmlNode) {
        
        TiXmlNode *connNode = xmlNode->FirstChild("Connection");
        
        if (connNode) {
            
            getAttributeString(connNode->ToElement(), "IP", serverIP, sizeof(serverIP));
            getAttributeInt(connNode->ToElement(), "Port", (int &) serverPort);
            char temp[100];
            
            if (getAttributeString(connNode->ToElement(), "UseThreads", temp, sizeof(temp))) {
                
                if (!strcasecmp(temp, "YES"))
                    useThreads = true;
                else if (!strcasecmp(temp, "NO"))
                    useThreads = false;
                else {
                    
#ifdef _LOGGER
                    Logger::logoutput(Logger::LOG_CONFIG, log_undefined_usethreads);
#endif
                    
                    cerr << "Undefined UseThreads value, possible values are YES/NO" << endl;
                }
            }
        }
        
        TiXmlNode *playerNode = xmlNode->FirstChild("Player");
        
        if (playerNode) {
            
            getAttributeString(playerNode->ToElement(), "Team", player.team, sizeof(player.team));
            getAttributeInt(playerNode->ToElement(), "Number", player.number);
        }
        
        TiXmlNode *worldNode = xmlNode->FirstChild("World");
        
        if (worldNode) {
            
            TiXmlNode *ballNode = worldNode->FirstChild("Ball");
            getAttributeDouble(ballNode->ToElement(), "Radius", world.ballRadius);
        }
        
        TiXmlNode *loggerNode = xmlNode->FirstChild("Logger");
        
        if (loggerNode) {
            
            TiXmlNode *moduleNode = loggerNode->FirstChild("Module");
            
            for (; moduleNode != NULL; moduleNode = moduleNode->NextSiblingElement()) {
                
                char temp[100];
                bool state;
                
                getAttributeString(moduleNode->ToElement(), "State", temp, sizeof(temp));
                
                if (!strcasecmp(temp, "on"))
                    state = true;
                else if (!strcasecmp(temp, "off"))
                    state = false;
                else {
                    
                    cerr << "Undefined State value, possible values are ON/OFF" << endl;
                    return 1;
                }
                
                getAttributeString(moduleNode->ToElement(), "Name", temp, sizeof(temp));
                
                for (int i = 0; i < Logger::LOG_COUNT; i++) {
                    
                    if (!strcasecmp(temp, Logger::getSwitchName(i)))
                        Logger::SetSwitch(i, state);
                }
            }
        }
    }
    
    return 0;
}
