/*! \file PlayerModel.cpp
 <pre>
 <b>File:</b> PlayerModel.cpp
 <b>Project:</b> Robocup 3D
 <b>Authors:</b> Hviezdna Jedenastka
 <b>Created:</b> 03/15/2008
 <b>Contents:</b> This class represents Player model = state of all players joints (their angles, speed, etc.) and gyroscope.
 </pre>
 */

#include "PlayerModel.h"
#include "Logger.h"


PlayerModel::PlayerModel() {
    
    gyroscope = new Gyroscope();
    
    FRP = new ForceResistance();
    
    ankleL = new UniversalJoint("lle5_6", JOINT_STOP_Lle5Min, JOINT_STOP_Lle5Max, JOINT_STOP_Lle6Min, JOINT_STOP_Lle6Max);
    ankleR = new UniversalJoint("rle5_6", JOINT_STOP_Rle5Min, JOINT_STOP_Rle5Max, JOINT_STOP_Rle6Min, JOINT_STOP_Rle6Max);
    
    kneeL = new HingeJoint("lle4", JOINT_STOP_Lle4Min, JOINT_STOP_Lle4Max);
    kneeR = new HingeJoint("rle4", JOINT_STOP_Rle4Min, JOINT_STOP_Rle4Max);
    
    legL = new HingeJoint("lle1", JOINT_STOP_Lle1Min, JOINT_STOP_Lle1Max);
    legR = new HingeJoint("rle1", JOINT_STOP_Rle1Min, JOINT_STOP_Rle1Max);
    
    hipL = new UniversalJoint("lle2_3", JOINT_STOP_Lle2Min, JOINT_STOP_Lle2Max, JOINT_STOP_Lle3Min, JOINT_STOP_Lle3Max);
    hipR = new UniversalJoint("rle2_3", JOINT_STOP_Rle2Min, JOINT_STOP_Rle2Max, JOINT_STOP_Rle3Min, JOINT_STOP_Rle3Max);
    
    shoulderL = new UniversalJoint("lae1_2", JOINT_STOP_Lae1Min, JOINT_STOP_Lae1Max, JOINT_STOP_Lae2Min, JOINT_STOP_Lae2Max);
    shoulderR = new UniversalJoint("rae1_2", JOINT_STOP_Rae1Min, JOINT_STOP_Rae1Max, JOINT_STOP_Rae2Min, JOINT_STOP_Rae2Max);
    
    elbowL = new HingeJoint("lae4", JOINT_STOP_Lae4Min, JOINT_STOP_Lae4Max);
    elbowR = new HingeJoint("rae4", JOINT_STOP_Rae4Min, JOINT_STOP_Rae4Max);
    
    armL = new HingeJoint("lae3", JOINT_STOP_Lae3Min, JOINT_STOP_Lae3Max);
    armR = new HingeJoint("rae3", JOINT_STOP_Rae3Min, JOINT_STOP_Rae3Max);
}

PlayerModel::~PlayerModel() {
    
    delete gyroscope;
    
    delete FRP;
    
    delete ankleL;
    delete ankleR;
    
    delete kneeL;
    delete kneeR;
    
    delete legL;
    delete legR;
    
    delete hipL;
    delete hipR;
    
    delete shoulderL;
    delete shoulderR;
    
    delete elbowL;
    delete elbowR;
    
    delete armL;
    delete armR;
}

void PlayerModel::update(ServerMsgParserData * parsedData){
    
    simTime = parsedData->GetSimTime();
    
    gyroscope->update(parsedData->GetGyroRate(), simTime);
    
    FRP->update(parsedData->GetFrpLocCoord_LF(), parsedData->GetFrpLocCoord_RF(),
                parsedData->GetFrpForceVector_LF(), parsedData->GetFrpForceVector_RF());
    
    ankleL->update(parsedData->GetUJValue_ax1(UJ_llj5_6), parsedData->GetUJValue_ax2(UJ_llj5_6), simTime);
    ankleR->update(parsedData->GetUJValue_ax1(UJ_rlj5_6), parsedData->GetUJValue_ax2(UJ_rlj5_6), simTime);
    
    kneeL->update(parsedData->GetHJValue(HJ_llj4), simTime);
    kneeR->update(parsedData->GetHJValue(HJ_rlj4), simTime);
    
    legL->update(parsedData->GetHJValue(HJ_llj1), simTime);
    legR->update(parsedData->GetHJValue(HJ_rlj1), simTime);
    
    hipL->update(parsedData->GetUJValue_ax1(UJ_llj2_3), parsedData->GetUJValue_ax2(UJ_llj2_3), simTime);
    hipR->update(parsedData->GetUJValue_ax1(UJ_rlj2_3), parsedData->GetUJValue_ax2(UJ_rlj2_3), simTime);
    
    shoulderL->update(parsedData->GetUJValue_ax1(UJ_laj1_2), parsedData->GetUJValue_ax2(UJ_laj1_2), simTime);
    shoulderR->update(parsedData->GetUJValue_ax1(UJ_raj1_2), parsedData->GetUJValue_ax2(UJ_raj1_2), simTime);
    
    elbowL->update(parsedData->GetHJValue(HJ_laj4), simTime);
    elbowR->update(parsedData->GetHJValue(HJ_raj4), simTime);
    
    armL->update(parsedData->GetHJValue(HJ_laj3), simTime);
    armR->update(parsedData->GetHJValue(HJ_raj3), simTime);
}

void PlayerModel::buildMessage(char *msg) {
    
    ankleL->buildMessage(msg);
    ankleR->buildMessage(msg + strlen(msg));
    
    kneeL->buildMessage(msg + strlen(msg));
    kneeR->buildMessage(msg + strlen(msg));
    
    legL->buildMessage(msg + strlen(msg));
    legR->buildMessage(msg + strlen(msg));
    
    hipL->buildMessage(msg + strlen(msg));
    hipR->buildMessage(msg + strlen(msg));
    
    shoulderL->buildMessage(msg + strlen(msg));
    shoulderR->buildMessage(msg + strlen(msg));
    
    elbowL->buildMessage(msg + strlen(msg));
    elbowR->buildMessage(msg + strlen(msg));
    
    armL->buildMessage(msg + strlen(msg));
    armR->buildMessage(msg + strlen(msg));
}

void PlayerModel::logPlayerModel() {
    
    Logger::logoutput(Logger::LOG_PLAYER_SIMTIME, log_simtime, simTime);
    Logger::logoutput(Logger::LOG_PLAYER_GYROSCOPE, log_playermodel_gyroscope, gyroscope->absolute.x, gyroscope->absolute.y, gyroscope->absolute.z);
    Logger::logoutput(Logger::LOG_PLAYER_FRP, log_playermodel_FRP,
                      FRP->LocCoord_LF.x, FRP->LocCoord_LF.y, FRP->LocCoord_LF.z,
                      FRP->LocCoord_RF.x, FRP->LocCoord_RF.y, FRP->LocCoord_RF.z,
                      FRP->ForceVector_LF.x, FRP->ForceVector_LF.y, FRP->ForceVector_LF.z,
                      FRP->ForceVector_RF.x, FRP->ForceVector_RF.y, FRP->ForceVector_RF.z);
    
    
    Logger::logoutput(Logger::LOG_PLAYER_ANKLES, log_playermodel_ankles, RAD2DEG(ankleL->getactAngleX()),
                      RAD2DEG(ankleL->getactAngleY()), RAD2DEG(ankleR->getactAngleX()), RAD2DEG(ankleR->getactAngleY()));
    Logger::logoutput(Logger::LOG_PLAYER_HIPS, log_playermodel_hips, RAD2DEG(hipL->getactAngleX()),
                      RAD2DEG(hipL->getactAngleY()), RAD2DEG(hipR->getactAngleX()), RAD2DEG(hipR->getactAngleY()));
    Logger::logoutput(Logger::LOG_PLAYER_SHOULDERS, log_playermodel_shoulders, RAD2DEG(shoulderL->getactAngleX()),
                      RAD2DEG(shoulderL->getactAngleY()), RAD2DEG(shoulderR->getactAngleX()), RAD2DEG(shoulderR->getactAngleY()));
    Logger::logoutput(Logger::LOG_PLAYER_KNEES, log_playermodel_knees, RAD2DEG(kneeL->getActAngleX()), RAD2DEG(kneeR->getActAngleX()));
    Logger::logoutput(Logger::LOG_PLAYER_LEGS, log_playermodel_legs, RAD2DEG(legL->getActAngleX()), RAD2DEG(legR->getActAngleX()));
    Logger::logoutput(Logger::LOG_PLAYER_ELBOWS, log_playermodel_elbows, RAD2DEG(elbowL->getActAngleX()), RAD2DEG(elbowR->getActAngleX()));
    Logger::logoutput(Logger::LOG_PLAYER_ARMS, log_playermodel_arms, RAD2DEG(armL->getActAngleX()), RAD2DEG(armR->getActAngleX()));
}

bool PlayerModel::jointsMoving() {
    
    return (ankleR->isMoving() || ankleL->isMoving() || kneeL->isMoving() || kneeR->isMoving() ||
            legL->isMoving() || legR->isMoving() || hipL->isMoving() || hipR->isMoving() ||
            shoulderL->isMoving() || shoulderR->isMoving() || elbowL->isMoving() || elbowR->isMoving() || armL->isMoving() || armR->isMoving());
}
