/*! \file HingeJoint.cpp
 <pre>
 <b>File:</b> HingeJoint.cpp
 <b>Project:</b> Robocup 3D
 <b>Authors:</b> Hviezdna Jedenastka
 <b>Created:</b> 03/15/2008
 <b>Contents:</b> This class extends from abstract class Joint.  This class represents simple type of Joint in PlayerModel.
                  (Knees, Legs, Elbows, Arms). Supports moving only in one dimension (X dimension).
 </pre>
 */

#include "HingeJoint.h"
#include <math.h>
#include "Logger.h"

HingeJoint::HingeJoint(char *_name, double _minStopDeg, double _maxStopDeg):minStopDeg(_minStopDeg), maxStopDeg(_maxStopDeg), Joint(_name) {
    
    actAngleX = 0.0;
    setAngleX = 0.0;
    speedX = 0.0;
    gainX = 0.0;
    movingX = false;
    timeoutReseted = false;
}

HingeJoint::~HingeJoint() {
}

void HingeJoint::recalculateConstant() {
    
	if (fabs(setAngleX - actAngleX) > JOINT_ANGLE_ERROR_CONSTANT) {
        
		speedX = gainX / (SIMULATION_STEP * 100.0);
		movingX = true;
		
	} else {
        
		speedX = 0.0;
		movingX = false;
	}
}

void HingeJoint::recalculateApproximating() {
    
    if (fabs(setAngleX - actAngleX) > JOINT_ANGLE_ERROR_APPROXIMATING) {
        
        speedX = gainX * (setAngleX - actAngleX) / (SIMULATION_STEP * 100.0);
        movingX = true;
        
    } else {
        
        speedX = 0.0;
        movingX = false;
    }
}

void HingeJoint::update(double _actAngleX, double _simTime) {
    
    Joint::update(_simTime);
    actAngleX = DEG2RAD(_actAngleX);
}

bool HingeJoint::moveTo(double _setAngleX, double _gainX, bool interceptMove, JointMoveMethod _method, double _timeout) {
    
    if (interceptMove || !movingX || timeoutReseted) {
        
		method = _method;
		startTime = simTime;
		timeout = _timeout;
		timeoutReseted = false;
        
		if (CHECK_JOINT_STOPS) {
            
			setFinalAngleChecked(_setAngleX);
            
		} else {
            
			setAngleX = DEG2RAD(_setAngleX);
		}
        
		gainX = _gainX;
    }
	
    // Must be recalculated here, because of movingX test
    recalculate();
    
    // Timeout check for move. Move returns true if timeout occurs, but joint
    // continues performing the move until timeout is reseted.
    if (simTime - startTime >= timeout) {
        
		return true;
    }
    
    return !movingX;
}

bool HingeJoint::moveBy(double _diffAngleX, double _gainX, bool interceptMove, JointMoveMethod _method, double _timeout) {
    
    return moveTo(RAD2DEG(actAngleX) + _diffAngleX, _gainX, interceptMove, _method, _timeout);
}

void HingeJoint::buildMessage(char *message) {
    
    sprintf(message, "(%s %f)", name, speedX);
}

void HingeJoint::setFinalAngleChecked(double _angleX) {
    
    if (_angleX < minStopDeg) {
        
        setAngleX = DEG2RAD(minStopDeg);
        
    } else if (_angleX > maxStopDeg) {
        
        setAngleX = DEG2RAD(maxStopDeg);
        
    } else {
        
        setAngleX = DEG2RAD(_angleX);
    }
}
