/*! \file HingeJoint.h
 <pre>
 <b>File:</b> HingeJoint.h
 <b>Project:</b> Robocup 3D
 <b>Authors:</b> Hviezdna Jedenastka
 <b>Created:</b> 03/15/2008
 <b>Contents:</b> This class extends from abstract class Joint.  This class represents simple type of Joint in PlayerModel.
                  (Knees, Legs, Elbows, Arms). Supports moving only in one dimension (X dimension).
 </pre>
 */

#ifndef __HINGEJOINT_H__
#define __HINGEJOINT_H__

#include "Joint.h"

/**
 * Hinge joint class.
 * Handles hinge joint calculations.
 */
class HingeJoint: public Joint {
    
public:
    
	double actAngleX;			/*!< actual joint angle (position) */
	double setAngleX;			/*!< final (desired) joint angle (position) */
	double gainX;				/*!< starting speed */
	
	bool movingX;				/*!< moving flag (while moving = true, when finished = false) */
	bool timeoutReseted;
	
	double minStopDeg;			/*!< minimum joint angle */
	double maxStopDeg;			/*!< maximum joint angle */
	
protected:
    
	/**
	 * Constant joint move calculation.
	 * Sets actual joint speed to value specified by gainX.
	 */
	void recalculateConstant();
	
	/**
	 * Approximating joint move calculation.
	 * Calculates difference between final angle and actual angle, multiplies this difference by gainX and sets it as actual joint speed.
	 */
	void recalculateApproximating();
	
	/**
	 * Keeps joint angle between its minimum and maximum limitations.
	 *
	 * @param	_angleX			joint angle
	 */
	void setFinalAngleChecked(double _angleX);	
    
public:
    
	double speedX;				/*!< actual joint speed */
	
	/**
	 * Constructor
	 *
	 * @param	_name			joint name
	 * @param	_minStopDeg		minimum joint angle
	 * @param	_maxStopDeg		maximum joint angle
	 */
	HingeJoint(char * _name, double _minStopDeg, double _maxStopDeg);
	
	/**
	 *	Destructor
	 */
	~HingeJoint();
    
	/**
	 * Joint update.
	 * Updades joint axis and simulation time.
	 *
	 * @param	_actAngleX		actual joint angle X
	 * @param	_simTime		actual simulation time
	 */
	void update(double _actAngleX, double _simTime);
    
	/**
	 * Absolute joint move.
	 * Moves joint to _setAngleX angle. It is absolute move. If CHECK_JOINT_STOPS (Joint.h) is set to true, joint limitations will be checked.
	 * 
	 * @param	_setAngleX		final angle to be achieved by move
	 * @param	_gainX			controls speed of the move
	 * @param	interceptMove	if set to false, started move will be finished, otherwise new final angle and move method will be used
	 * @param	_method			move method
	 * @param	_timeout		maximum joint move time
	 *
	 * @return	true if final angle has been achieved within the error range specified by _JOINTANGLEERROR<method> constant in Joint.h
	 */
	bool moveTo(double _setAngleX, double _gainX, bool interceptMove, JointMoveMethod _method, double _timeout);
    
	/**
	 * Relative joint move.
	 * Moves joint by _diffAngleX. It uses moveTo method to perform relative move.
	 * 
	 * @param	_diffAngleX		relative angle
	 * @param	_gainX			controls speed of the move
	 * @param	interceptMove	if set to false, started move will be finished, otherwise new final angle and move method will be used
	 * @param	_method			move method
	 * @param	_timeout		maximum joint move time
	 *
	 * @return	true if joint angle has been changed by _diffAngle within the error range specified by _JOINTANGLEERROR<method> constant in Joint.h
	 */
	bool moveBy(double _diffAngleX, double _gainX, bool interceptMove, JointMoveMethod _method, double _timeout);
    
	/**
	 * Builds joint message.
	 *
	 * @param	message			pointer to the message buffer
	 */
	void buildMessage(char * message);
	
	/**
	 * Gets actual joint angle.
	 *
	 * @return actual joint angle
	 */
	double getActAngleX() {
        
		return actAngleX;
	}
	
	/**
	 * Tests joint move.
	 *
	 * @return true if joint is moving
	 */
	bool isMoving() {
        
		return (movingX);
	}
	
	bool resetTimeout() {
        
		timeoutReseted = true;
		return true;
	}
};

#endif /*__HINGEJOINT_H__*/
