/*! \file UniversalJoint.cpp
 <pre>
 <b>File:</b> UniversalJoint.cpp
 <b>Project:</b> Robocup 3D
 <b>Authors:</b> Hviezdna Jedenastka
 <b>Created:</b> 03/15/2008
 <b>Contents:</b> This class extends from abstract class Joint. This class represents Universal type of Joint in PlayerModel.
                  (Ankles, Hips, Shoulders). Supports moving in two dimensions (X and Y).
 </pre>
 */

#include "UniversalJoint.h"

using namespace std;

UniversalJoint::UniversalJoint(char * _name, double _minStopDegX, double _maxStopDegX, double _minStopDegY, double _maxStopDegY): Joint(_name), minStopDegX(_minStopDegX), maxStopDegX(_maxStopDegX), minStopDegY(_minStopDegY), maxStopDegY(_maxStopDegY) {
    
	actAngleX	= 0.0;
	actAngleY	= 0.0;
	setAngleX	= 0.0;
	setAngleY	= 0.0;
	speedX		= 0.0;
	speedY		= 0.0;
	gainX		= 0.0;
	gainY		= 0.0;
	movingX		= false;
	movingY		= false;
	timeoutReseted	= false;
}

UniversalJoint::~UniversalJoint() {
}

void UniversalJoint::recalculateConstant() {
    
	// X axis calculation
	if (ABS(setAngleX - actAngleX) > JOINT_ANGLE_ERROR_CONSTANT) {
        
		speedX = gainX / (SIMULATION_STEP * 100.0);
		movingX = true;
        
	} else {
        
		speedX = 0.0;
		movingX = false;
	}
    
	// Y axis calculation
	if (ABS(setAngleY - actAngleY) > JOINT_ANGLE_ERROR_CONSTANT) {
        
		speedY = gainY / (SIMULATION_STEP * 100.0);
		movingY = true;
		
	} else {
        
		speedY = 0.0;
		movingY = false;
	}
}

void UniversalJoint::recalculateApproximating() {
    
	// X axis calculation
	// While current angle isn't close enough, approximate final angle.
	// Otherwise stop moving joint in X axis.
	if (ABS(setAngleX - actAngleX) > JOINT_ANGLE_ERROR_APPROXIMATING) {
        
		speedX = gainX * (setAngleX - actAngleX) / (SIMULATION_STEP * 100.0);
		movingX = true;
        
	} else {
        
		speedX = 0.0;
		movingX = false;
	}
    
	// Y axis calculation
	// Similar to X axis calculation.
	if (ABS(setAngleY - actAngleY) > JOINT_ANGLE_ERROR_APPROXIMATING) {
        
		speedY = gainY * (setAngleY - actAngleY) / (SIMULATION_STEP * 100.0);
		movingY = true;
		
	} else {
        
		speedY = 0.0;
		movingY = false;
	}
}

void UniversalJoint::update(double _actAngleX, double _actAngleY, double _simTime) {
	
	Joint::update(_simTime);
    
	actAngleX = DEG2RAD(_actAngleX);
	actAngleY = DEG2RAD(_actAngleY);
}

bool UniversalJoint::moveTo(double _setAngleX, double _gainX, double _setAngleY, double _gainY, bool interceptMove, JointMoveMethod _method, double _timeout) {
    
	if (interceptMove || (!movingX && !movingY) || timeoutReseted) {
        
		method = _method;
		startTime = simTime;
		timeout = _timeout;
		timeoutReseted = false;
        
		if (CHECK_JOINT_STOPS) {
            
			setFinalAnglesChecked(_setAngleX, _setAngleY);
			
		} else {
            
			setAngleX = DEG2RAD(_setAngleX);
			setAngleY = DEG2RAD(_setAngleY);
		}
        
		gainX = _gainX;
		gainY = _gainY;
	}
	
	// Must be recalculated here, because of movingX and movingY test
	recalculate();
    
	// Timeout check for move. Move returns true if timeout occurs, but joint
	// continues performing the move.
	if (simTime - startTime >= timeout) {
        
		return true;
	}
    
	return (!movingX && !movingY);
}

bool UniversalJoint::moveBy(double _setAngleX, double _gainX, double _setAngleY, double _gainY, bool interceptMove, JointMoveMethod _method, double _timeout) {
    
	return moveTo(RAD2DEG(actAngleX) + _setAngleX, _gainX, RAD2DEG(actAngleY) + _setAngleY, _gainY, interceptMove, _method, _timeout);
}

void UniversalJoint::buildMessage(char * message) {
    
	sprintf(message, "(%s %f %f)", name, speedX, speedY);
}

void UniversalJoint::setFinalAnglesChecked(double _angleX, double _angleY) {
	
	// X axis	
	if (_angleX < minStopDegX) {
        
		setAngleX = DEG2RAD(minStopDegX);
		
	} else if (_angleX > maxStopDegX) {
        
		setAngleX = DEG2RAD(maxStopDegX);
        
	} else {
        
		setAngleX = DEG2RAD(_angleX);
	}
    
	// Y axis
	if (_angleY < minStopDegY) {
        
		setAngleY	= DEG2RAD(minStopDegY);
        
	} else if (_angleY > maxStopDegY) {
        
		setAngleY = DEG2RAD(maxStopDegY);
		
	} else {
        
		setAngleY	= DEG2RAD(_angleY);
	}
}