/*! \file UniversalJoint.h
 <pre>
 <b>File:</b> UniversalJoint.h
 <b>Project:</b> Robocup 3D
 <b>Authors:</b> Hviezdna Jedenastka
 <b>Created:</b> 03/15/2008
 <b>Contents:</b> This class extends from abstract class Joint. This class represents Universal type of Joint in PlayerModel.
                  (Ankles, Hips, Shoulders). Supports moving in two dimensions (X and Y).
 </pre>
 */

#ifndef __UNIVERSALJOINT_H__
#define __UNIVERSALJOINT_H__

#include "Joint.h"

/**
 * Universal joint class.
 * Handles universal joint calculations.
 */
class UniversalJoint: public Joint {
    
private:
    
	double actAngleX;		/*!< actual X dimension joint angle (position) */
	double actAngleY;		/*!< actual Y dimension joint angle (position) */
	double setAngleX;		/*!< final (desired) X dimension joint angle (position) */
	double setAngleY;		/*!< final (desired) Y dimension joint angle (position) */
	double gainX;			/*!< starting X dimension speed */
	double gainY;			/*!< starting Y dimension speed */
	
	bool movingX;			/*!< moving X flag (while moving in X dimension = true, when finished = false) */
	bool movingY;			/*!< moving Y flag (while moving in Y dimension = true, when finished = false) */
	bool timeoutReseted;
    
	double minStopDegX;		/*!< minimum joint angle X */
	double maxStopDegX;		/*!< maximum joint angle X */
	double minStopDegY;		/*!< minimum joint angle Y */
	double maxStopDegY;		/*!< maximum joint angle Y */
    
protected:
    
	/**
	 * Constant joint move calculation.
	 * Sets actual joint speed to value specified by gainX.
	 */
	void recalculateConstant();
    
	/**
	 * Approximating joint move calculation.
	 * Calculates difference between final angle and actual angle, multiplies this difference by gainX and sets it as actual joint speed.
	 */
	void recalculateApproximating();
    
	/**
	 * Keeps joint angle between its minimum and maximum limitations.
	 *
	 * @param	_angleX			joint angle X
	 * @param	_angleY			joint angle Y
	 */
	void setFinalAnglesChecked(double _angleX, double _angleY);
	
public:
	
	double speedX;			/*!< actual joint speed X */
	double speedY;			/*!< actual joint speed Y */
    
	/**
	 * Constructor
	 *
	 * @param	_name			joint name
	 * @param	_minStopDegX	minimum joint angle X
	 * @param	_maxStopDegX	maximum joint angle X
	 * @param	_minStopDegY	minimum joint angle Y
	 * @param	_maxStopDegY	maximum joint angle Y
	 */
	UniversalJoint(char * _name, double _minStopDegX, double _maxStopDegX, double _minStopDegY, double _maxStopDegY);
    
	/**
	 *	Destructor
	 */
	~UniversalJoint();
    
	/**
	 * Joint update.
	 * Updades joint axes and simulation time.
	 *
	 * @param	_actAngleX		actual joint angle X
	 * @param	_actAngleY		actual joint angle Y
	 * @param	_simTime		actual simulation time
	 */
	void update(double _actAngleX, double _actAngleY, double _simTime);
    
	/**
	 * Absolute joint move.
	 * Moves joint to _setAngleX and _setAngleY angle. It is absolute move. If CHECK_JOINT_STOPS (Joint.h) is set to true, joint limitations will be checked.
	 * 
	 * @param	_setAngleX		final angle to be achieved by move on X axis
	 * @param	_gainX			controls speed of the move on X axis according to move method
	 * @param	_setAngleY		final angle to be achieved by move on Y axis
	 * @param	_gainY			controls speed of the move on Y axis according to move method
	 * @param	interceptMove	if set to false, started move will be finished, otherwise new final angle and move method will be used
	 * @param	_method			move method
	 * @param	_timeout		maximum joint move time
	 *
	 * @return	true if final angle has been achieved within the error range specified by _JOINTANGLEERROR<method> constant in Joint.h
	 */
	bool moveTo(double _setAngleX, double _gainX, double _setAngleY, double _gainY, bool interceptMove, JointMoveMethod _method, double _timeout);
    
	/**
	 * Relative joint move.
	 * Moves joint by _diffAngleX and _diffAngleY. It uses moveTo method to perform relative move.
	 * 
	 * @param	_diffAngleX		relative angle on X axis
	 * @param	_gainX			controls speed of the move on X axis according to move method
	 * @param	_diffAngleY		relative angle on Y axis
	 * @param	_gainY			controls speed of the move on Y axis according to move method
	 * @param	interceptMove	if set to false, started move will be finished, otherwise new final angle and move method will be used
	 * @param	_method			move method
	 * @param	_timeout		maximum joint move time
	 *
	 * @return	true if joint angle has been changed by _diffAngle within the error range specified by _JOINTANGLEERROR<method> constant in Joint.h
	 */
	bool moveBy(double _diffAngleX, double _gainX, double _diffAngleY, double _gainY, bool interceptMove, JointMoveMethod _method, double _timeout);
    
	/**
	 * Builds joint message.
	 *
	 * @param	message			pointer to the message buffer
	 */
	void buildMessage(char * message);
    
	/**
	 * Gets actual joint angle X.
	 *
	 * @return actual joint angle X
	 */
	double getactAngleX() {
        
		return actAngleX;
	}
	
	/**
	 * Gets actual joint angle Y.
	 *
	 * @return actual joint angle Y
	 */
	double getactAngleY() {
        
		return actAngleY;
	}
	
	/**
	 * Tests joint move.
	 *
	 * @return true if joint is moving at least in one axis
	 */
	bool isMoving() {
        
		return (movingX || movingY);
	}
	
	bool resetTimeout() {
        
		timeoutReseted = true;
		return true;
	}
};

#endif /*__UNIVERSALJOINT_H__*/
